#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

from pathlib import Path
from typing import Any, Callable, Optional, Tuple, Union

import pandas as pd
from torchvision.datasets.vision import VisionDataset

from health_cpath.datasets.base_dataset import TilesDataset
from health_cpath.models.transforms import load_pil_image

from health_cpath.datasets.dataset_return_index import DatasetWithReturnIndex


class PandaTilesDataset(TilesDataset):
    """
    Dataset class for loading PANDA tiles.

    Iterating over this dataset returns a dictionary containing:
    - `'slide_id'` (str): parent slide ID (`'image_id'` in the PANDA dataset)
    - `'tile_id'` (str)
    - `'image'` (`PIL.Image`): RGB tile
    - `'mask'` (str): path to mask PNG file
    - `'tile_x'`, `'tile_y'` (int): top-right tile coordinates
    - `'data_provider'`, `'slide_isup_grade'`, `'slide_gleason_score'` (str): parent slide metadata
    """
    SPLIT_COLUMN = None  # PANDA does not have an official train/test split

    def __init__(self,
                 root: Path,
                 dataset_csv: Optional[Union[str, Path]] = None,
                 dataset_df: Optional[pd.DataFrame] = None,
                 occupancy_threshold: Optional[float] = None,
                 random_subset_fraction: Optional[float] = None,
                 label_column: str = "slide_isup_grade",
                 n_classes: int = 6) -> None:
        """
        :param root: Root directory of the dataset.
        :param dataset_csv: Full path to a dataset CSV file, containing at least
        `TILE_ID_COLUMN`, `SLIDE_ID_COLUMN`, and `IMAGE_COLUMN`. If omitted, the CSV will be read
        from `"{root}/{DEFAULT_CSV_FILENAME}"`.
        :param dataset_df: A potentially pre-processed dataframe in the same format as would be read
        from the dataset CSV file, e.g. after some filtering. If given, overrides `dataset_csv`.
        :occupancy_threshold: A value between 0-1 such that only tiles with occupancy > occupancy_threshold
        will be selected. If 0, all tiles are selected. If `None` (default), all tiles are selected.
        :random_subset_fraction: A value > 0 and <=1 such that this proportion of tiles will be randomly selected.
        If 1, all tiles are selected. If `None` (default), all tiles are selected.
        """
        super().__init__(root=Path(root),
                         dataset_csv=dataset_csv,
                         dataset_df=dataset_df,
                         train=None,
                         validate_columns=False,
                         label_column=label_column,
                         n_classes=n_classes)

        if occupancy_threshold is not None:
            if (occupancy_threshold < 0) or (occupancy_threshold > 1):
                raise ValueError(f"Occupancy threshold value {occupancy_threshold} should be in range 0-1.")
            dataset_df_filtered = self.dataset_df.loc[
                self.dataset_df['occupancy'] > occupancy_threshold
            ]
            self.dataset_df = dataset_df_filtered

        if random_subset_fraction is not None:
            if (random_subset_fraction <= 0) or (random_subset_fraction > 1):
                raise ValueError(f"Random subset fraction value {random_subset_fraction} should be > 0 and < = 1.")
            df_length_random_subset_fraction = round(len(self.dataset_df) * random_subset_fraction)
            dataset_df_filtered = self.dataset_df.sample(n=df_length_random_subset_fraction)
            self.dataset_df = dataset_df_filtered

        self.copy_coordinates_columns()
        self.validate_columns()


class PandaTilesDatasetReturnImageLabel(VisionDataset):
    """
    Any dataset used in SSL needs to return a tuple where the first element is the image and the second is a
    class label.
    """
    occupancy_threshold = 0
    random_subset_fraction = 1

    def __init__(self,
                 root: Path,
                 dataset_csv: Optional[Union[str, Path]] = None,
                 dataset_df: Optional[pd.DataFrame] = None,
                 transform: Optional[Callable] = None,
                 **kwargs: Any) -> None:
        super().__init__(root=root, transform=transform)

        self.base_dataset = PandaTilesDataset(root=root,
                                              dataset_csv=dataset_csv,
                                              dataset_df=dataset_df,
                                              occupancy_threshold=self.occupancy_threshold,
                                              random_subset_fraction=self.random_subset_fraction)

    def __getitem__(self, index: int) -> Tuple:  # type: ignore
        sample = self.base_dataset[index]
        # TODO change to a meaningful evaluation
        image = load_pil_image(sample[self.base_dataset.IMAGE_COLUMN])
        if self.transform:
            image = self.transform(image)
        # get binary label
        label = 0 if sample[self.base_dataset.label_column] == 0 else 1
        return image, label

    def __len__(self) -> int:
        return len(self.base_dataset)


class PandaTilesDatasetWithReturnIndex(DatasetWithReturnIndex, PandaTilesDatasetReturnImageLabel):
    """
    Any dataset used in SSL needs to inherit from DatasetWithReturnIndex as well as VisionData.
    This class is just a shorthand notation for this double inheritance. Please note that this class needs
    to override __getitem__(), this is why we need a separate PandaTilesDatasetReturnImageLabel.
    """
    @property
    def num_classes(self) -> int:
        return 2



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from typing import Tuple

from health_ml.utils.split_dataset import DatasetSplits

from health_cpath.datamodules.base_module import SlidesDataModule, TilesDataModule
from health_cpath.datasets.panda_dataset import PandaDataset
from health_cpath.datasets.panda_tiles_dataset import PandaTilesDataset


class PandaTilesDataModule(TilesDataModule):
    """ PandaTilesDataModule is the child class of TilesDataModule specific to PANDA dataset
    Method get_splits() returns the train, val, test splits from the PANDA dataset
    """

    def get_splits(self) -> Tuple[PandaTilesDataset, PandaTilesDataset, PandaTilesDataset]:
        dataset = PandaTilesDataset(self.root_path)
        splits = DatasetSplits.from_proportions(dataset.dataset_df.reset_index(),
                                                proportion_train=.8,
                                                proportion_test=.1,
                                                proportion_val=.1,
                                                subject_column=dataset.TILE_ID_COLUMN,
                                                group_column=dataset.SLIDE_ID_COLUMN)

        if self.crossval_count > 1:
            # Function get_k_fold_cross_validation_splits() will concatenate train and val splits
            splits = splits.get_k_fold_cross_validation_splits(self.crossval_count)[self.crossval_index]

        return (PandaTilesDataset(self.root_path, dataset_df=splits.train),
                PandaTilesDataset(self.root_path, dataset_df=splits.val),
                PandaTilesDataset(self.root_path, dataset_df=splits.test))


class PandaSlidesDataModule(SlidesDataModule):
    """ PandaSlidesDataModule is the child class of SlidesDataModule specific to PANDA dataset
    Method get_splits() returns the train, val, test splits from the PANDA dataset
    """

    def get_splits(self) -> Tuple[PandaDataset, PandaDataset, PandaDataset]:
        dataset = PandaDataset(self.root_path)
        splits = DatasetSplits.from_proportions(dataset.dataset_df.reset_index(),
                                                proportion_train=.8,
                                                proportion_test=.1,
                                                proportion_val=.1,
                                                subject_column=dataset.SLIDE_ID_COLUMN)

        if self.crossval_count > 1:
            # Function get_k_fold_cross_validation_splits() will concatenate train and val splits
            splits = splits.get_k_fold_cross_validation_splits(self.crossval_count)[self.crossval_index]

        return (PandaDataset(self.root_path, dataset_df=splits.train),
                PandaDataset(self.root_path, dataset_df=splits.val),
                PandaDataset(self.root_path, dataset_df=splits.test))



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

from typing import Tuple

from health_ml.utils.split_dataset import DatasetSplits

from health_cpath.datamodules.base_module import TilesDataModule
from health_cpath.datasets.tcga_crck_tiles_dataset import TcgaCrck_TilesDataset


class TcgaCrckTilesDataModule(TilesDataModule):
    """ TcgaCrckTilesDataModule is the child class of TilesDataModule specific to TCGA-Crck dataset
    """

    def get_splits(self) -> Tuple[TcgaCrck_TilesDataset, TcgaCrck_TilesDataset, TcgaCrck_TilesDataset]:
        """
        :return: the train, val, test splits from the TCGA-Crck dataset
        """
        trainval_dataset = TcgaCrck_TilesDataset(self.root_path, train=True)
        splits = DatasetSplits.from_proportions(trainval_dataset.dataset_df.reset_index(),
                                                proportion_train=0.8,
                                                proportion_test=0.0,
                                                proportion_val=0.2,
                                                subject_column=trainval_dataset.TILE_ID_COLUMN,
                                                group_column=trainval_dataset.SLIDE_ID_COLUMN,
                                                random_seed=5)

        if self.crossval_count > 1:
            # Function get_k_fold_cross_validation_splits() will concatenate train and val splits
            splits = splits.get_k_fold_cross_validation_splits(self.crossval_count)[self.crossval_index]

        return (TcgaCrck_TilesDataset(self.root_path, dataset_df=splits.train),
                TcgaCrck_TilesDataset(self.root_path, dataset_df=splits.val),
                TcgaCrck_TilesDataset(self.root_path, train=False))



#  -------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  -------------------------------------------------------------------------------------------
from pathlib import Path

import pandas as pd
from health_cpath.datasets.default_paths import TCGA_CRCK_DATASET_ID

from health_cpath.utils.tcga_utils import extract_fields
from health_cpath.datasets.tcga_prad_dataset import TcgaPradDataset


def check_dataset_csv_paths(dataset_dir: Path) -> None:
    df = pd.read_csv(dataset_dir / TcgaPradDataset.DEFAULT_CSV_FILENAME)
    for img_path in df.image:
        assert (dataset_dir / img_path).is_file()


if __name__ == '__main__':
    # Script needs to be started in the parent folder of the dataset folder
    current_dir = Path.cwd()
    expected_datasetdir = TCGA_CRCK_DATASET_ID
    if not (current_dir / expected_datasetdir).is_dir:
        raise ValueError(f"The current folder must contain the actual dataset folder {expected_datasetdir}")
    dataset_dir = current_dir / expected_datasetdir
    expected_subdirs = ["CRC_DX_TEST", "CRC_DX_TRAIN"]
    if not all([(dataset_dir / subdir).is_dir() for subdir in expected_subdirs]):
        raise ValueError(f"The folder {expected_datasetdir} needs to have these subfolder: {expected_subdirs}")
    image_paths = [str(image_path.relative_to(dataset_dir))
                   for split_dir in dataset_dir.iterdir()
                   for class_dir in split_dir.iterdir()
                   for image_path in class_dir.iterdir()]

    df = pd.DataFrame(image_paths, columns=['image'])

    # takes up to ~20 seconds
    df = df.apply(extract_fields, axis='columns', result_type='expand')
    df.to_csv(dataset_dir / TcgaPradDataset.DEFAULT_CSV_FILENAME, index=False)

    check_dataset_csv_paths(dataset_dir)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import param
import numpy as np
import skimage.filters

from enum import Enum
from health_ml.utils import box_utils
from health_cpath.utils.naming import SlideKey
from monai.data.wsi_reader import WSIReader
from monai.transforms import MapTransform, LoadImaged
from typing import Any, Callable, Dict, Optional, Tuple
from health_azure.logging import print_message_with_rank_pid


def get_luminance(slide: np.ndarray) -> np.ndarray:
    """Compute a grayscale version of the input slide.

    :param slide: The RGB image array in (*, C, H, W) format.
    :return: The single-channel luminance array as (*, H, W).
    """
    # TODO: Consider more sophisticated luminance calculation if necessary
    return slide.mean(axis=-3)  # type: ignore


def segment_foreground(slide: np.ndarray, threshold: Optional[float] = None) -> Tuple[np.ndarray, float]:
    """Segment the given slide by thresholding its luminance.

    :param slide: The RGB image array in (*, C, H, W) format.
    :param threshold: Pixels with luminance below this value will be considered foreground.
    If `None` (default), an optimal threshold will be estimated automatically using Otsu's method.
    :return: A tuple containing the boolean output array in (*, H, W) format and the threshold used.
    """
    luminance = get_luminance(slide)
    if threshold is None:
        threshold = skimage.filters.threshold_otsu(luminance)
    return luminance < threshold, threshold


class ROIType(str, Enum):
    """Options for the ROI selection. Either a bounding box defined by foreground or a mask can be used."""
    FOREGROUND = 'foreground'
    MASK = 'segmentation_mask'
    WHOLE = 'whole_slide'


class WSIBackend(str, Enum):
    """Options for the WSI reader backend."""
    OPENSLIDE = 'OpenSlide'
    CUCIM = 'cuCIM'


class BaseLoadROId:
    """Abstract base class for loading a region of interest (ROI) from a slide. The ROI is defined by a bounding box."""

    def __init__(
        self, backend: str = WSIBackend.CUCIM, image_key: str = SlideKey.IMAGE, level: int = 1, margin: int = 0,
        backend_args: Dict = {}
    ) -> None:
        """
        :param backend: The WSI reader backend to use. One of 'OpenSlide' or 'cuCIM'. Default: 'cuCIM'.
        :param image_key: Image key in the input and output dictionaries. Default: 'image'.
        :param level: Magnification level to load from the raw multi-scale file, 0 is the highest resolution. Default: 1
            which loads the second highest resolution.
        :param margin: Amount in pixels by which to enlarge the estimated bounding box for cropping. Default: 0.
        :param backend_args: Additional arguments to pass to the WSI reader backend. Default: {}.
        """
        self.reader = WSIReader(backend=backend, **backend_args)
        self.image_key = image_key
        self.level = level
        self.margin = margin

    def _get_foreground_mask(self, slide_obj: Any, level: int) -> np.ndarray:
        """Estimate foreground mask at the given level of the slide."""
        raise NotImplementedError

    def _get_whole_slide_bbox(self, slide_obj: Any, level: int) -> box_utils.Box:
        """Return a bounding box that covers the whole slide at the given level."""
        h, w = self.reader.get_size(slide_obj, level=level)
        return box_utils.Box(0, 0, w, h)

    def _get_bounding_box(self, slide_obj: Any, slide_id: int) -> box_utils.Box:
        """Estimate bounding box at the lowest resolution (i.e. highest level) of the slide."""
        highest_level = self.reader.get_level_count(slide_obj) - 1
        scale = self.reader.get_downsample_ratio(slide_obj, highest_level)
        foreground_mask = self._get_foreground_mask(slide_obj, level=highest_level)
        try:
            bbox = box_utils.get_bounding_box(foreground_mask)
        except RuntimeError as e:
            print_message_with_rank_pid(f"Failed to estimate bounding box for slide {slide_id}: {e}")
            bbox = self._get_whole_slide_bbox(slide_obj, level=highest_level)
        return scale * bbox.add_margin(self.margin)

    def __call__(self, data: Dict) -> Dict:
        raise NotImplementedError


class LoadROId(MapTransform, BaseLoadROId):
    """Transform that loads a pathology slide, cropped to an estimated bounding box (ROI) of the foreground tissue.

    Operates on dictionaries, replacing the file path in `image_key` with the loaded array in
    (C, H, W) format. Also adds the following entries:
    - `SlideKey.ORIGIN` (tuple): top-right coordinates of the bounding box
    - `SlideKey.SCALE` (float): corresponding scale, loaded from the file
    - `SlideKey.FOREGROUND_THRESHOLD` (float): threshold used to segment the foreground
    """

    def __init__(
        self, image_key: str = SlideKey.IMAGE, foreground_threshold: Optional[float] = None, **kwargs: Any
    ) -> None:
        """
        :param image_key: Image key in the input and output dictionaries. Default: 'image'.
        :param foreground_threshold: Pixels with luminance below this value will be considered foreground.
        If `None` (default), an optimal threshold will be estimated automatically using Otsu's method.
        :param kwargs: Additional arguments for `BaseLoadROId`.
        """
        MapTransform.__init__(self, [image_key], allow_missing_keys=False)
        BaseLoadROId.__init__(self, image_key=image_key, **kwargs)
        self.foreground_threshold = foreground_threshold

    def _get_foreground_mask(self, slide_obj: Any, level: int) -> np.ndarray:
        """Estimate foreground mask at the highest resolution (i.e. lowest level) of the slide based on luminance."""
        slide, _ = self.reader.get_data(slide_obj, level=level)
        foreground_mask, threshold = segment_foreground(slide, self.foreground_threshold)
        self.foreground_threshold = threshold
        return foreground_mask

    def __call__(self, data: Dict) -> Dict:
        try:
            image_obj = self.reader.read(data[self.image_key])
            level0_bbox = self._get_bounding_box(image_obj, data[SlideKey.SLIDE_ID])

            # cuCIM/OpenSlide takes absolute location coordinates in the level 0 reference frame,
            # but relative region size in pixels at the chosen level
            origin = (level0_bbox.y, level0_bbox.x)
            scale = self.reader.get_downsample_ratio(image_obj, self.level)
            scaled_bbox = level0_bbox / scale

            data[self.image_key], _ = self.reader.get_data(image_obj, location=origin, level=self.level,
                                                           size=(scaled_bbox.h, scaled_bbox.w))
            data[SlideKey.ORIGIN] = origin
            data[SlideKey.SCALE] = scale
            data[SlideKey.FOREGROUND_THRESHOLD] = self.foreground_threshold
        finally:
            image_obj.close()
        return data


class LoadMaskROId(MapTransform, BaseLoadROId):
    """Transform that loads a pathology slide and mask, cropped to the mask bounding box (ROI) defined by the mask.

    Operates on dictionaries, replacing the file paths in `image_key` and `mask_key` with the
    respective loaded arrays, in (C, H, W) format. Also adds the following meta-data entries:
    - `'location'` (tuple): top-right coordinates of the bounding box
    - `'size'` (tuple): width and height of the bounding box
    - `'level'` (int): chosen magnification level
    - `'scale'` (float): corresponding scale, loaded from the file
    """

    def __init__(self, image_key: str = SlideKey.IMAGE, mask_key: str = SlideKey.MASK, **kwargs: Any) -> None:
        """
        :param image_key: Image key in the input and output dictionaries. Default: 'image'.
        :param mask_key: Mask key in the input and output dictionaries. Default: 'mask'.
        :param kwargs: Additional arguments for `BaseLoadROId`.
        """
        MapTransform.__init__(self, [image_key, mask_key], allow_missing_keys=False)
        BaseLoadROId.__init__(self, image_key=image_key, **kwargs)
        self.mask_key = mask_key

    def _get_foreground_mask(self, mask_obj: Any, level: int) -> np.ndarray:
        """Load foreground mask at the given level of the slide."""
        mask, _ = self.reader.get_data(mask_obj, level=level)  # loaded as RGB PIL image
        foreground_mask = mask[0] > 0  # PANDA segmentation mask is in 'R' channel
        return foreground_mask

    def __call__(self, data: Dict) -> Dict:
        try:
            mask_obj = self.reader.read(data[self.mask_key])
            image_obj = self.reader.read(data[self.image_key])

            level0_bbox = self._get_bounding_box(mask_obj, data[SlideKey.SLIDE_ID])

            # cuCIM/OpenSlide take absolute location coordinates in the level 0 reference frame,
            # but relative region size in pixels at the chosen level
            scale = self.reader.get_downsample_ratio(mask_obj, self.level)
            scaled_bbox = level0_bbox / scale
            origin = (level0_bbox.y, level0_bbox.x)
            get_data_kwargs = dict(
                location=origin,
                size=(scaled_bbox.h, scaled_bbox.w),
                level=self.level,
            )
            mask, _ = self.reader.get_data(mask_obj, **get_data_kwargs)  # type: ignore
            data[self.mask_key] = mask[:1]  # PANDA segmentation mask is in 'R' channel
            data[self.image_key], _ = self.reader.get_data(image_obj, **get_data_kwargs)  # type: ignore
            data.update(get_data_kwargs)
            data[SlideKey.SCALE] = scale
            data[SlideKey.ORIGIN] = origin
        finally:
            mask_obj.close()
            image_obj.close()
        return data


class LoadingParams(param.Parameterized):
    """Parameters for loading a whole slide image."""

    level: int = param.Integer(
        default=1,
        doc="Magnification level to load from the raw multi-scale files. Default: 1.")
    margin: int = param.Integer(
        default=0, doc="Amount in pixels by which to enlarge the estimated bounding box for cropping")
    backend: WSIBackend = param.ClassSelector(
        default=WSIBackend.CUCIM,
        class_=WSIBackend,
        doc="WSI reader backend. Default: cuCIM.")
    roi_type: ROIType = param.ClassSelector(
        default=ROIType.WHOLE,
        class_=ROIType,
        doc="ROI type to use for cropping the slide. Default: `ROIType.WHOLE`. no cropping is performed.")
    image_key: str = param.String(
        default=SlideKey.IMAGE,
        doc="Key for the image in the data dictionary.")
    mask_key: str = param.String(
        default=SlideKey.MASK,
        doc="Key for the mask in the data dictionary. This only applies to `LoadMaskROId`.")
    foreground_threshold: Optional[float] = param.Number(
        default=None,
        bounds=(0, 255.),
        allow_None=True,
        doc="Threshold for foreground mask. If None, the threshold is selected automatically with otsu thresholding."
        "This only applies to `LoadROId`.")

    def set_roi_type_to_foreground(self) -> None:
        """Set the ROI type to foreground. This is useful for plotting even if we load whole slides during
        training. This help us reduce the size of thrumbnails to only meaningful tissue. We only hardcode it to
        foreground in the WHOLE case. Otherwise, keep it as is if a mask is available."""
        if self.roi_type == ROIType.WHOLE:
            self.roi_type = ROIType.FOREGROUND

    def get_load_roid_transform(self) -> Callable:
        """Returns a transform to load a slide and mask, cropped to the mask bounding box (ROI) defined by either the
        mask or the foreground."""
        if self.roi_type == ROIType.WHOLE:
            return LoadImaged(keys=self.image_key, reader=WSIReader, image_only=True, level=self.level,  # type: ignore
                              backend=self.backend, dtype=np.uint8, **self.get_additionl_backend_args())
        elif self.roi_type == ROIType.FOREGROUND:
            return LoadROId(backend=self.backend, image_key=self.image_key, level=self.level,
                            margin=self.margin, foreground_threshold=self.foreground_threshold,
                            backend_args=self.get_additionl_backend_args())
        elif self.roi_type == ROIType.MASK:
            return LoadMaskROId(backend=self.backend, image_key=self.image_key,
                                mask_key=self.mask_key, level=self.level, margin=self.margin,
                                backend_args=self.get_additionl_backend_args())
        else:
            raise ValueError(f"Unknown ROI type: {self.roi_type}. Choose from {list(ROIType)}.")

    def get_additionl_backend_args(self) -> Dict[str, Any]:
        """Returns a dictionary of additional arguments for the WSI reader backend. Multi processing is
        enabled since monai 1.0.0 by specifying num_workers > 0 with CuCIM backend only.
        This function can be overridden in BaseMIL to add additional arguments for the backend."""
        return dict()

    def should_upscale_coordinates(self) -> bool:
        """Returns True if the coordinates should be upscaled to the level 0 reference frame. This is the case when
        we load the whole slide without cropping for tiling on the fly. We need to upscale the coordinates to the
        highest level to match the coordinates of the tiles pipeline."""
        return self.level > 0 and self.roi_type == ROIType.WHOLE



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import pandas as pd
from pathlib import Path
from typing import Any, Dict, Union, Optional
from health_cpath.datasets.base_dataset import SlidesDataset


class PandaDataset(SlidesDataset):
    """Dataset class for loading files from the PANDA challenge dataset.

    Iterating over this dataset returns a dictionary following the `SlideKey` schema plus meta-data
    from the original dataset (`'data_provider'`, `'isup_grade'`, and `'gleason_score'`).

    Ref.: https://www.kaggle.com/c/prostate-cancer-grade-assessment/overview
    """
    SLIDE_ID_COLUMN = 'image_id'
    IMAGE_COLUMN = 'image'
    MASK_COLUMN = 'mask'

    METADATA_COLUMNS = ('data_provider', 'isup_grade', 'gleason_score')

    DEFAULT_CSV_FILENAME = "train.csv"

    def __init__(self,
                 root: Union[str, Path],
                 dataset_csv: Optional[Union[str, Path]] = None,
                 dataset_df: Optional[pd.DataFrame] = None,
                 label_column: str = "isup_grade",
                 n_classes: int = 6,
                 dataframe_kwargs: Dict[str, Any] = {}) -> None:
        super().__init__(root, dataset_csv, dataset_df, validate_columns=False, label_column=label_column,
                         n_classes=n_classes, dataframe_kwargs=dataframe_kwargs)
        # PANDA CSV does not come with paths for image and mask files
        slide_ids = self.dataset_df.index
        self.dataset_df[self.IMAGE_COLUMN] = "train_images/" + slide_ids + ".tiff"
        self.dataset_df[self.MASK_COLUMN] = "train_label_masks/" + slide_ids + "_mask.tiff"
        self.validate_columns()



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

from enum import Enum


class SlideKey(str, Enum):
    SLIDE_ID = 'slide_id'
    IMAGE = 'image'
    IMAGE_PATH = 'image_path'
    MASK = 'mask'
    MASK_PATH = 'mask_path'
    LABEL = 'label'
    SPLIT = 'split'
    SCALE = 'scale'
    ORIGIN = 'origin'
    FOREGROUND_THRESHOLD = 'foreground_threshold'
    METADATA = 'metadata'
    LOCATION = 'location'


class TileKey(str, Enum):
    TILE_ID = 'tile_id'
    SLIDE_ID = 'slide_id'
    IMAGE = 'image'
    IMAGE_PATH = 'image_path'
    MASK = 'mask'
    MASK_PATH = 'mask_path'
    LABEL = 'label'
    SPLIT = 'split'
    TILE_LEFT = 'left'
    TILE_TOP = 'top'
    TILE_RIGHT = 'right'
    TILE_BOTTOM = 'bottom'
    OCCUPANCY = 'occupancy'
    FOREGROUND_THRESHOLD = 'foreground_threshold'
    SLIDE_METADATA = 'slide_metadata'
    NUM_DISCARDED = 'num_discarded'

    @staticmethod
    def from_slide_metadata_key(slide_metadata_key: str) -> str:
        return 'slide_' + slide_metadata_key


class ResultsKey(str, Enum):
    SLIDE_ID = 'slide_id'
    TILE_ID = 'tile_id'
    FEATURES = 'features'
    IMAGE_PATH = 'image_path'
    LOSS = 'loss'
    LOSS_PER_SAMPLE = 'loss_per_sample'
    PROB = 'prob'
    CLASS_PROBS = 'prob_class'
    PRED_LABEL = 'pred_label'
    TRUE_LABEL = 'true_label'
    BAG_ATTN = 'bag_attn'
    TILE_LEFT = 'left'
    TILE_TOP = 'top'
    TILE_RIGHT = 'right'
    TILE_BOTTOM = 'bottom'
    ENTROPY = 'entropy'


class MetricsKey(str, Enum):
    ACC = 'accuracy'
    ACC_MACRO = 'macro_accuracy'
    ACC_WEIGHTED = 'weighted_accuracy'
    CONF_MATRIX = 'confusion_matrix'
    AUROC = 'auroc'
    PRECISION = 'precision'
    RECALL = 'recall'
    F1 = 'f1score'
    COHENKAPPA = 'cohenkappa'
    AVERAGE_PRECISION = 'average_precision'
    SPECIFICITY = 'specificity'


class ModelKey(str, Enum):
    TRAIN = 'train'
    VAL = 'val'
    TEST = 'test'


class AMLMetricsJsonKey(str, Enum):
    HYPERPARAMS = 'hyperparams'
    NAME = 'name'
    VALUE = 'value'
    N_CLASSES = 'n_classes'
    CLASS_NAMES = 'class_names'
    MAX_EPOCHS = 'max_epochs'


class PlotOption(Enum):
    TOP_BOTTOM_TILES = "top_bottom_tiles"
    SLIDE_THUMBNAIL = "slide_thumbnail"
    ATTENTION_HEATMAP = "attention_heatmap"
    ATTENTION_HISTOGRAM = "attention_histogram"
    CONFUSION_MATRIX = "confusion_matrix"
    HISTOGRAM = "histogram"
    PR_CURVE = "pr_curve"


class DeepMILSubmodules(str, Enum):
    ENCODER = 'encoder'
    POOLING = 'aggregation_fn'
    CLASSIFIER = 'classifier_fn'



#  -------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  -------------------------------------------------------------------------------------------

from pathlib import Path
from typing import Dict, List, Optional

import pandas as pd
from matplotlib import pyplot as plt

from health_azure.utils import get_aml_run_from_run_id, get_workspace
from health_ml.utils.reports import HTMLReport
from health_cpath.utils.analysis_plot_utils import (add_training_curves_legend, plot_confusion_matrices,
                                                    plot_hyperdrive_roc_and_pr_curves,
                                                    plot_hyperdrive_training_curves)
from health_cpath.utils.output_utils import (AML_LEGACY_TEST_OUTPUTS_CSV, AML_TEST_OUTPUTS_CSV,
                                             AML_VAL_OUTPUTS_CSV)
from health_cpath.utils.report_utils import (collect_hyperdrive_metrics, collect_hyperdrive_outputs,
                                             child_runs_have_val_and_test_outputs, get_best_epoch_metrics,
                                             get_best_epochs, get_child_runs_hyperparams, get_hyperdrive_metrics_table,
                                             get_formatted_run_info, collect_class_info, get_max_epochs,
                                             download_hyperdrive_metrics_if_required)
from health_cpath.utils.naming import MetricsKey, ModelKey


def generate_html_report(parent_run_id: str, output_dir: Path,
                         workspace_config_path: Optional[Path] = None,
                         include_test: bool = False, overwrite: bool = False,
                         hyperdrive_arg_name: str = "crossval_index",
                         primary_metric: str = MetricsKey.AUROC) -> None:
    """
    Function to generate an HTML report of a Hyperdrive AML run (e.g., cross validation, different random seeds, ...).

    :param run_id: The parent Hyperdrive run ID.
    :param output_dir: Directory where to download Azure ML data and save the report.
    :param workspace_config_path: Path to Azure ML workspace config.json file.
        If omitted, will try to load default workspace.
    :param include_test: Include test results in the generated report.
    :param overwrite: Forces (re)download of metrics and output files, even if they already exist locally.
    :param hyperdrive_arg_name: Name of the Hyperdrive argument used for indexing the child runs.
        Default `crossval_index`.
    :param primary_metric: Name of the reference metric to optimise. Default `MetricsKey.AUROC`.
    """
    aml_workspace = get_workspace(workspace_config_path=workspace_config_path)
    parent_run = get_aml_run_from_run_id(parent_run_id, aml_workspace=aml_workspace)
    report_dir = output_dir / parent_run.display_name
    report_dir.mkdir(parents=True, exist_ok=True)

    report = HTMLReport(output_folder=report_dir)

    report.add_text(get_formatted_run_info(parent_run))

    report.add_heading("Azure ML metrics", level=2)

    # Download metrics from AML. Can take several seconds for each child run
    metrics_json = download_hyperdrive_metrics_if_required(parent_run_id, report_dir, aml_workspace,
                                                           overwrite=overwrite, hyperdrive_arg_name=hyperdrive_arg_name)

    # Get metrics dataframe from the downloaded json file
    metrics_df = collect_hyperdrive_metrics(metrics_json=metrics_json)

    hyperparameters_children = get_child_runs_hyperparams(metrics_df)
    max_epochs_dict = get_max_epochs(hyperparams_children=hyperparameters_children)
    best_epochs = get_best_epochs(metrics_df=metrics_df, primary_metric=f'{ModelKey.VAL}/{primary_metric}',
                                  max_epochs_dict=max_epochs_dict, maximise=True)

    # Add training curves for loss and AUROC (train and val.)
    render_training_curves(report, heading="Training curves", level=3,
                           metrics_df=metrics_df, best_epochs=best_epochs, report_dir=report_dir,
                           primary_metric=primary_metric)

    # Get metrics list with class names
    num_classes, class_names = collect_class_info(hyperparams_children=hyperparameters_children)

    base_metrics_list: List[str] = [MetricsKey.ACC, MetricsKey.AUROC, MetricsKey.AVERAGE_PRECISION,
                                    MetricsKey.COHENKAPPA]
    if num_classes > 1:
        base_metrics_list += [MetricsKey.ACC_MACRO, MetricsKey.ACC_WEIGHTED]
    else:
        base_metrics_list += [MetricsKey.F1, MetricsKey.PRECISION, MetricsKey.RECALL, MetricsKey.SPECIFICITY]

    base_metrics_list += class_names

    # Add tables with relevant metrics (val. and test)
    render_metrics_table(report,
                         heading=f"Validation metrics (best epoch based on maximum validation {primary_metric})",
                         level=3,
                         metrics_df=metrics_df, best_epochs=best_epochs,
                         base_metrics_list=base_metrics_list, metrics_prefix=f'{ModelKey.VAL}/')

    if include_test:
        render_metrics_table(report, heading="Test metrics", level=3,
                             metrics_df=metrics_df, best_epochs=None,
                             base_metrics_list=base_metrics_list, metrics_prefix=f'{ModelKey.TEST}/')

    # Get output data frames if available
    try:
        has_val_and_test_outputs = child_runs_have_val_and_test_outputs(parent_run)
        if has_val_and_test_outputs:
            output_filename_val = AML_VAL_OUTPUTS_CSV
            outputs_dfs_val = collect_hyperdrive_outputs(parent_run_id=parent_run_id, download_dir=report_dir,
                                                         aml_workspace=aml_workspace,
                                                         output_filename=output_filename_val, overwrite=overwrite,
                                                         hyperdrive_arg_name=hyperdrive_arg_name)
            if include_test:
                output_filename_test = AML_TEST_OUTPUTS_CSV if has_val_and_test_outputs else AML_LEGACY_TEST_OUTPUTS_CSV
                outputs_dfs_test = collect_hyperdrive_outputs(parent_run_id=parent_run_id, download_dir=report_dir,
                                                              aml_workspace=aml_workspace,
                                                              output_filename=output_filename_test, overwrite=overwrite,
                                                              hyperdrive_arg_name=hyperdrive_arg_name)

        if num_classes == 1:
            # Currently ROC and PR curves rendered only for binary case
            # TODO: Enable rendering of multi-class ROC and PR curves
            report.add_heading("ROC and PR curves", level=2)
            if has_val_and_test_outputs:
                # Add val. ROC and PR curves
                render_roc_and_pr_curves(report=report, heading="Validation ROC and PR curves", level=3,
                                         report_dir=report_dir,
                                         outputs_dfs=outputs_dfs_val,
                                         prefix=f'{ModelKey.VAL}_')
            if include_test:
                # Add test ROC and PR curves
                render_roc_and_pr_curves(report=report, heading="Test ROC and PR curves", level=3,
                                         report_dir=report_dir,
                                         outputs_dfs=outputs_dfs_test,
                                         prefix=f'{ModelKey.TEST}_')

        # Add confusion matrices for each fold
        report.add_heading("Confusion matrices", level=2)
        if has_val_and_test_outputs:
            # Add val. confusion matrices
            render_confusion_matrices(report=report, heading="Validation confusion matrices", level=3,
                                      class_names=class_names,
                                      report_dir=report_dir, outputs_dfs=outputs_dfs_val,
                                      prefix=f'{ModelKey.VAL}_')

        if include_test:
            # Add test confusion matrices
            render_confusion_matrices(report=report, heading="Test confusion matrices", level=3,
                                      class_names=class_names,
                                      report_dir=report_dir, outputs_dfs=outputs_dfs_test,
                                      prefix=f'{ModelKey.TEST}_')

    except ValueError as e:
        print(e)
        print("Since all expected output files were not found, skipping ROC-PR curves and confusion matrices.")

    # TODO: Add qualitative model outputs
    # report.add_heading("Qualitative model outputs", level=2)

    print(f"Rendering report to: {report.report_path_html.resolve()}")
    report.render()


def render_training_curves(report: HTMLReport, heading: str, level: int,
                           metrics_df: pd.DataFrame, best_epochs: Optional[Dict[int, int]],
                           report_dir: Path, primary_metric: str = MetricsKey.AUROC) -> None:
    """
    Function to render training curves for HTML reports.

    :param report: HTML report to perform the rendering.
    :param heading: Heading of the section.
    :param level: Level of HTML heading (e.g. sub-section, sub-sub-section) corresponding to HTML heading levels.
    :param metrics_df: Metrics dataframe, as returned by :py:func:`collect_hyperdrive_metrics()` and
        :py:func:`~health_azure.aggregate_hyperdrive_metrics()`.
    :param best_epochs: Dictionary mapping each hyperdrive child index to its best epoch.
    :param report_dir: Directory of the HTML report.
    :param primary_metric: Primary metric name. Default is AUROC.
    """
    report.add_heading(heading, level=level)
    metrics = {"loss_epoch", MetricsKey.AUROC.value, primary_metric}
    fig, axs = plt.subplots(1, len(metrics), figsize=(5 * len(metrics), 4))
    for i, metric in enumerate(metrics):
        plot_hyperdrive_training_curves(metrics_df, train_metric=f'{ModelKey.TRAIN}/{metric}',
                                        val_metric=f'{ModelKey.VAL}/{metric}',
                                        ylabel=metric, best_epochs=best_epochs, ax=axs[i])
    add_training_curves_legend(fig, include_best_epoch=True)
    training_curves_fig_path = report_dir / "training_curves.png"
    fig.savefig(training_curves_fig_path, bbox_inches='tight')
    report.add_images([training_curves_fig_path], base64_encode=True)


def render_metrics_table(report: HTMLReport, heading: str, level: int,
                         metrics_df: pd.DataFrame, best_epochs: Optional[Dict[int, int]],
                         base_metrics_list: List[str], metrics_prefix: str) -> None:
    """
    Function to render metrics table for HTML reports.

    :param report: HTML report to perform the rendering.
    :param heading: Heading of the section.
    :param level: Level of HTML heading (e.g. sub-section, sub-sub-section) corresponding to HTML heading levels.
    :param metrics_df: Metrics dataframe, as returned by :py:func:`collect_hyperdrive_metrics()` and
        :py:func:`~health_azure.aggregate_hyperdrive_metrics()`.
    :param base_metrics_list: List of metric names to include in the table.
    :param best_epochs: Dictionary mapping each hyperdrive child index to its best epoch.
    :param metrics_prefix: Prefix to add to the metrics names (e.g. `val`, `test`)
    """
    report.add_heading(heading, level=level)
    metrics_list = [metrics_prefix + metric for metric in base_metrics_list]
    if best_epochs:
        metrics_df = get_best_epoch_metrics(metrics_df, metrics_list, best_epochs)
    metrics_table = get_hyperdrive_metrics_table(metrics_df, metrics_list)
    report.add_tables([metrics_table])


def render_roc_and_pr_curves(report: HTMLReport, heading: str, level: int, report_dir: Path,
                             outputs_dfs: Dict[int, pd.DataFrame], prefix: str = '') -> None:
    """
    Function to render ROC and PR curves for HTML reports.

    :param report: HTML report to perform the rendering.
    :param heading: Heading of the section.
    :param level: Level of HTML heading (e.g. sub-section, sub-sub-section) corresponding to HTML heading levels.
    :param report_dir: Local directory where the report is stored.
    :param outputs_dfs: A dictionary of dataframes with the sorted hyperdrive child runs indices as keys.
    :param prefix: Prefix to add to the figures saved (e.g. `val`, `test`).
    """
    report.add_heading(heading, level=level)
    fig = plot_hyperdrive_roc_and_pr_curves(outputs_dfs, scores_column='prob_class1')
    roc_pr_curves_fig_path = report_dir / f"{prefix}roc_pr_curves.png"
    fig.savefig(roc_pr_curves_fig_path, bbox_inches='tight')
    report.add_images([roc_pr_curves_fig_path], base64_encode=True)


def render_confusion_matrices(report: HTMLReport, heading: str, level: int, class_names: List[str],
                              report_dir: Path, outputs_dfs: Dict[int, pd.DataFrame], prefix: str = '') -> None:
    """
    Function to render confusion matrices for HTML reports.

    :param report: HTML report to perform the rendering.
    :param heading: Heading of the section.
    :param level: Level of HTML heading (e.g. sub-section, sub-sub-section) corresponding to HTML heading levels.
    :param class_names: Names of classes.
    :param report_dir: Local directory where the report is stored.
    :param outputs_dfs: A dictionary of dataframes with the sorted hyperdrive child runs indices as keys.
    :param prefix: Prefix to add to the figures saved (e.g. `val`, `test`).
    """
    report.add_heading(heading, level=level)
    fig = plot_confusion_matrices(hyperdrive_dfs=outputs_dfs, class_names=class_names)
    confusion_matrices_fig_path = report_dir / f"{prefix}confusion_matrices.png"
    fig.savefig(confusion_matrices_fig_path, bbox_inches='tight')
    report.add_images([confusion_matrices_fig_path], base64_encode=True)


if __name__ == "__main__":
    """
    Usage example from CLI:
    python generate_hyperdrive_report.py \
    --run_id <insert AML run ID here> \
    --output_dir outputs \
    --include_test
    """

    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument('--run_id', help="The parent Hyperdrive run ID.")
    parser.add_argument('--output_dir', help="Directory where to download Azure ML data and save the report.")
    parser.add_argument('--workspace_config', help="Path to Azure ML workspace config.json file. "
                                                   "If omitted, will try to load default workspace.")
    parser.add_argument('--include_test', action='store_true', help="Opt-in flag to include test results "
                                                                    "in the generated report.")
    parser.add_argument('--overwrite', action='store_true', help="Forces (re)download of metrics and output files, "
                                                                 "even if they already exist locally.")
    parser.add_argument("--hyper_arg_name", default="crossval_index",
                        help="Name of the Hyperdrive argument used for indexing the child runs.")
    parser.add_argument("--primary_metric", default=MetricsKey.AUROC, help="Name of the reference metric to optimise.")
    args = parser.parse_args()

    if args.output_dir is None:
        args.output_dir = Path.cwd() / "outputs"
    workspace_config = Path(args.workspace_config).resolve() if args.workspace_config else None

    print(f"Output dir: {Path(args.output_dir).resolve()}")
    if workspace_config is not None:
        if not workspace_config.is_file():
            raise ValueError(f"Specified workspace config file does not exist: {workspace_config}")
        print(f"Workspace config: {workspace_config}")

    generate_html_report(parent_run_id=args.run_id,
                         output_dir=Path(args.output_dir),
                         workspace_config_path=workspace_config,
                         include_test=args.include_test,
                         overwrite=args.overwrite,
                         hyperdrive_arg_name=args.hyper_arg_name,
                         primary_metric=args.primary_metric)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

"""DeepSMILECrck is the container for experiments relating to DeepSMILE using the TCGA-CRCk dataset.

For convenience, this module also defines encoder-specific containers e.g.TcgaCrckImageNetMIL

Reference:
- Schirris (2021). DeepSMILE: Self-supervised heterogeneity-aware multiple instance learning for DNA
damage response defect classification directly from H&E whole-slide images. arXiv:2107.09405
"""
from typing import Any, Set

from health_ml.networks.layers.attention_layers import AttentionLayer
from health_cpath.configs.run_ids import innereye_ssl_checkpoint_crck_4ws
from health_cpath.datamodules.base_module import TilesDataModule
from health_cpath.datamodules.tcga_crck_module import TcgaCrckTilesDataModule
from health_cpath.datasets.default_paths import TCGA_CRCK_DATASET_ID
from health_cpath.models.encoders import (
    HistoSSLEncoder,
    ImageNetSimCLREncoder,
    Resnet18,
    SSLEncoder,
)
from health_cpath.configs.classification.BaseMIL import BaseMILTiles
from health_cpath.datasets.tcga_crck_tiles_dataset import TcgaCrck_TilesDataset
from health_cpath.utils.naming import PlotOption
from health_ml.utils.checkpoint_utils import CheckpointParser


class DeepSMILECrck(BaseMILTiles):
    def __init__(self, **kwargs: Any) -> None:
        # Define dictionary with default params that can be overridden from subclasses or CLI
        default_kwargs = dict(
            # declared in BaseMIL:
            pool_type=AttentionLayer.__name__,
            num_transformer_pool_layers=4,
            num_transformer_pool_heads=4,
            encoding_chunk_size=60,
            tune_encoder=False,
            is_caching=True,
            num_top_slides=0,
            azure_datasets=[TCGA_CRCK_DATASET_ID],
            # declared in TrainerParams:
            max_epochs=50,
            # declared in WorkflowParams:
            crossval_count=5,
            # declared in OptimizerParams:
            l_rate=5e-4,
            weight_decay=1e-4,
            adam_betas=(0.9, 0.99),
        )
        default_kwargs.update(kwargs)
        super().__init__(**default_kwargs)

    def setup(self) -> None:
        super().setup()

    def get_data_module(self) -> TilesDataModule:
        return TcgaCrckTilesDataModule(
            root_path=self.local_datasets[0],
            max_bag_size=self.max_bag_size,
            batch_size=self.batch_size,
            batch_size_inf=self.batch_size_inf,
            max_bag_size_inf=self.max_bag_size_inf,
            transforms_dict=self.get_transforms_dict(TcgaCrck_TilesDataset.IMAGE_COLUMN),
            cache_mode=self.cache_mode,
            precache_location=self.precache_location,
            cache_dir=self.cache_dir,
            crossval_count=self.crossval_count,
            crossval_index=self.crossval_index,
            dataloader_kwargs=self.get_dataloader_kwargs(),
            seed=self.get_effective_random_seed(),
            pl_replace_sampler_ddp=self.pl_replace_sampler_ddp,
        )

    def get_test_plot_options(self) -> Set[PlotOption]:
        plot_options = super().get_test_plot_options()
        plot_options.add(PlotOption.PR_CURVE)
        return plot_options


class TcgaCrckImageNetMIL(DeepSMILECrck):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=Resnet18.__name__, **kwargs)


class TcgaCrckImageNetSimCLRMIL(DeepSMILECrck):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=ImageNetSimCLREncoder.__name__, **kwargs)


class TcgaCrckSSLMIL(DeepSMILECrck):
    def __init__(self, **kwargs: Any) -> None:
        # If no SSL checkpoint is provided, use the default one
        self.ssl_checkpoint = self.ssl_checkpoint or CheckpointParser(innereye_ssl_checkpoint_crck_4ws)
        super().__init__(encoder_type=SSLEncoder.__name__, **kwargs)


class TcgaCrckHistoSSLMIL(DeepSMILECrck):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=HistoSSLEncoder.__name__, **kwargs)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from health_ml.utils import health_ml_package_setup


def health_cpath_package_setup() -> None:
    """
    Set up the Python packages where needed. In particular, reduce the logging level for some of the used
    libraries, which are particularly talkative in DEBUG mode. Usually when running in DEBUG mode, we want
    diagnostics about the model building itself, but not for the underlying libraries.

    It also adds workarounds for known issues in some packages.
    """
    health_ml_package_setup()



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from typing import Any, Tuple, Union

import torch

OptionalIndexInputAndLabel = Union[Tuple[torch.Tensor, int], Tuple[int, torch.Tensor, int]]


class DatasetWithReturnIndex:
    """
    Class to be use with double inheritance with a VisionDataset.
    Overloads the __getitem__ function so that we can optionally also return
    the index within the dataset.
    """

    def __init__(self, root: str, return_index: bool, **kwargs: Any) -> None:
        self.return_index = return_index
        super().__init__(root=root, **kwargs)  # type: ignore

    def __getitem__(self, index: int) -> Any:
        item = super().__getitem__(index)  # type: ignore
        if self.return_index:
            return (index, *item)
        else:
            return item

    @property
    def num_classes(self) -> int:
        raise NotImplementedError



#  -------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  -------------------------------------------------------------------------------------------

from copy import deepcopy
import shutil
from itertools import chain
from pathlib import Path
from typing import Any, Collection, Dict, List, Mapping, Optional, Sequence, Tuple

import numpy as np
import pandas as pd
import torch
import logging

from ruamel.yaml import YAML
from torchmetrics.metric import Metric

from health_azure.utils import replace_directory
from health_cpath.datasets.base_dataset import SlidesDataset
from health_cpath.preprocessing.loading import LoadingParams
from health_cpath.utils.plots_utils import DeepMILPlotsHandler, TilesSelector
from health_cpath.utils.naming import MetricsKey, ModelKey, PlotOption, ResultsKey

OUTPUTS_CSV_FILENAME = "test_output.csv"
VAL_OUTPUTS_SUBDIR = "val"
PREV_VAL_OUTPUTS_SUBDIR = "val_old"
TEST_OUTPUTS_SUBDIR = "test"
EXTRA_VAL_OUTPUTS_SUBDIR = "extra_val"
EXTRA_PREFIX = "extra_"

AML_OUTPUTS_DIR = "outputs"
AML_LEGACY_TEST_OUTPUTS_CSV = "/".join([AML_OUTPUTS_DIR, OUTPUTS_CSV_FILENAME])
AML_VAL_OUTPUTS_CSV = "/".join([AML_OUTPUTS_DIR, VAL_OUTPUTS_SUBDIR, OUTPUTS_CSV_FILENAME])
AML_TEST_OUTPUTS_CSV = "/".join([AML_OUTPUTS_DIR, TEST_OUTPUTS_SUBDIR, OUTPUTS_CSV_FILENAME])

BatchResultsType = Dict[ResultsKey, Any]
EpochResultsType = List[BatchResultsType]
ResultsType = Dict[ResultsKey, List[Any]]


def validate_class_names(class_names: Optional[Sequence[str]], n_classes: int) -> Tuple[str, ...]:
    """Return valid names for the specified number of classes.

    :param class_names: List of class names. If `None`, will return `('0', '1', ...)`.
    :param n_classes: Number of classes. If `1` (binary), expects `len(class_names) == 2`.
    :return: Validated class names tuple with length `2` for binary classes (`n_classes == 1`), otherwise `n_classes`.
    """
    effective_n_classes = n_classes if n_classes > 1 else 2
    if class_names is None:
        class_names = [str(i) for i in range(effective_n_classes)]
    if len(class_names) != effective_n_classes:
        raise ValueError(f"Mismatch in number of class names ({class_names}) and number"
                         f"of classes ({effective_n_classes})")
    return tuple(class_names)


def validate_slide_datasets_for_plot_options(
    plot_options: Collection[PlotOption], slides_dataset: Optional[SlidesDataset]
) -> None:
    """Validate that the specified plot options are compatible with the specified slides dataset.

    :param plot_options: Plot options to validate.
    :param slides_dataset: Slides dataset to validate against.
    """

    def _validate_slide_plot_option(plot_option: PlotOption) -> None:
        if plot_option in plot_options and not slides_dataset:
            raise ValueError(f"Plot option {plot_option} requires a slides dataset")

    _validate_slide_plot_option(PlotOption.SLIDE_THUMBNAIL)
    _validate_slide_plot_option(PlotOption.ATTENTION_HEATMAP)


def normalize_dict_for_df(dict_old: Dict[ResultsKey, Any]) -> Dict[str, Any]:
    # slide-level dictionaries are processed by making value dimensions uniform and converting to numpy arrays.
    # these steps are required to convert the dictionary to pandas dataframe.
    dict_new: Dict[str, Any] = dict()
    bag_size = len(dict_old[ResultsKey.SLIDE_ID])
    for key, value in dict_old.items():
        if key not in [ResultsKey.CLASS_PROBS, ResultsKey.PROB]:
            if isinstance(value, torch.Tensor):
                value = value.squeeze(0).cpu().numpy()
                if value.ndim == 0:
                    value = np.full(bag_size, fill_value=value)
            dict_new[key] = value
        elif key == ResultsKey.CLASS_PROBS:
            if isinstance(value, torch.Tensor):
                value = value.squeeze(0).cpu().numpy()
                for i in range(len(value)):
                    dict_new[key + str(i)] = np.repeat(value[i], bag_size)
    return dict_new


def gather_results(epoch_results: EpochResultsType) -> EpochResultsType:
    """Gather epoch results across all DDP processes into a single list.

    :param epoch_results: The collected epoch results (i.e. list of batch results) for the current process.
    :return: A list in the same format as `epoch_results`, containing batch results from all DDP processes. Returns
        `epoch_results` unchanged if not in distributed mode.
    """
    if torch.distributed.is_initialized():
        world_size = torch.distributed.get_world_size()
        if world_size > 1:
            object_list: EpochResultsType = [None] * world_size  # type: ignore
            torch.distributed.all_gather_object(object_list, epoch_results)
            epoch_results = list(chain(*object_list))  # type: ignore
    return epoch_results


def collate_results_on_cpu(epoch_results: EpochResultsType) -> ResultsType:
    """Convert a list of results dictionaries into a dictionary of lists, with all tensors on CPU.

    :param epoch_results: Collected epoch results, whose elements are dictionaries of :py:class:`ResultsKey` to the
        outputs of the respective batch (i.e. indexed as `epoch_results[batch_index][results_key]`).
    :return: A dictionary mapping each :py:class:`ResultsKey` to a list containing the corresponding outputs for every
        batch (i.e. indexed as `collated_results[results_key][batch_index]`). All tensors will have been placed on CPU.
    """
    results: ResultsType = {}
    for key in epoch_results[0].keys():
        results[key] = []
        for batch_results in epoch_results:
            batch_elements = batch_results[key]
            if key == ResultsKey.LOSS:
                batch_elements = [batch_elements]
            batch_elements = [elem.cpu() if isinstance(elem, torch.Tensor) else elem
                              for elem in batch_elements]
            results[key].extend(batch_elements)
    return results


def save_outputs_csv(results: ResultsType, outputs_dir: Path) -> None:
    logging.info("Saving outputs ...")
    # collate at slide level
    list_slide_dicts: List[Dict[ResultsKey, Any]] = []
    # any column can be used here, the assumption is that the first dimension is the N of slides
    for slide_idx in range(len(results[ResultsKey.SLIDE_ID])):
        slide_dict = {key: results[key][slide_idx] for key in results
                      if key not in [ResultsKey.FEATURES, ResultsKey.LOSS]}
        list_slide_dicts.append(slide_dict)

    assert outputs_dir.is_dir(), f"No such dir: {outputs_dir}"
    logging.info(f"Metrics results will be output to {outputs_dir}")
    csv_filename = outputs_dir / OUTPUTS_CSV_FILENAME

    # Collect the list of dictionaries in a list of pandas dataframe and save
    df_list = []
    for slide_dict in list_slide_dicts:
        slide_dict = normalize_dict_for_df(slide_dict)  # type: ignore
        df_list.append(pd.DataFrame.from_dict(slide_dict))
    df = pd.concat(df_list, ignore_index=True)
    df.to_csv(csv_filename, mode='w+', header=True)


def save_features(results: ResultsType, outputs_dir: Path) -> None:
    # Collect all features in a list and save
    features_list = [features.squeeze(0).cpu() for features in results[ResultsKey.FEATURES]]
    torch.save(features_list, outputs_dir / 'test_encoded_features.pickle')


class OutputsPolicy:
    """Utility class that defines when to save validation epoch outputs."""

    _BEST_EPOCH_KEY = 'best_epoch'
    _BEST_VALUE_KEY = 'best_value'
    _PRIMARY_METRIC_KEY = 'primary_metric'

    def __init__(self, outputs_root: Path, primary_val_metric: MetricsKey, maximise: bool) -> None:
        """
        :param outputs_root: Root directory where to save a recovery file with best epoch and metric value.
        :param primary_val_metric: Name of the validation metric to track for saving best epoch outputs.
        :param maximise: Whether higher is better for `primary_val_metric`.
        """
        self.outputs_root = outputs_root
        self.primary_val_metric = primary_val_metric
        self.maximise = maximise

        self._init_best_metric()

    @property
    def best_metric_file_path(self) -> Path:
        return self.outputs_root / "best_val_metric.yml"

    def _init_best_metric(self) -> None:
        """Initialise running best metric epoch and value (recovered from disk if available).

        :raises ValueError: If the primary metric name does not match the one saved on disk.
        """
        if self.best_metric_file_path.exists():
            contents = YAML().load(self.best_metric_file_path)
            self._best_metric_epoch = contents[self._BEST_EPOCH_KEY]
            self._best_metric_value = contents[self._BEST_VALUE_KEY]
            if contents[self._PRIMARY_METRIC_KEY] != self.primary_val_metric:
                raise ValueError(f"Expected primary metric '{self.primary_val_metric}', but found "
                                 f"'{contents[self._PRIMARY_METRIC_KEY]}' in {self.best_metric_file_path}")
        else:
            self._best_metric_epoch = 0
            self._best_metric_value = float('-inf') if self.maximise else float('inf')

    def _save_best_metric(self) -> None:
        """Save best metric epoch, value, and name to disk, to allow recovery (e.g. in case of pre-emption)."""
        contents = {self._BEST_EPOCH_KEY: self._best_metric_epoch,
                    self._BEST_VALUE_KEY: self._best_metric_value,
                    self._PRIMARY_METRIC_KEY: self.primary_val_metric.value}
        YAML().dump(contents, self.best_metric_file_path)

    def should_save_validation_outputs(self, metrics_dict: Mapping[MetricsKey, Metric], epoch: int,
                                       is_global_rank_zero: bool = True, on_extra_val: bool = False) -> bool:
        """Determine whether validation outputs should be saved given the current epoch's metrics.

        :param metrics_dict: Current epoch's metrics dictionary from
            :py:class:`~health_cpath.models.deepmil.DeepMILModule`.
        :param epoch: Current epoch number.
        :param is_global_rank_zero: Whether this is the global rank-0 process in distributed scenarios.
            Set to `True` (default) if running a single process.
        :param on_extra_val: Whether this is an extra validation epoch (e.g. after training).
        :return: Whether this is the best validation epoch so far.
        """
        if on_extra_val:
            return False
        metric = metrics_dict[self.primary_val_metric]
        # If the metric hasn't been updated we don't want to save it
        if not metric._update_called:
            logging.warning("Encountered metric that hasn't been updated. Not saving.")
            return False
        # The metric needs to be computed on all ranks to allow synchronisation
        metric_value = float(metric.compute())

        # Validation outputs and best metric should be saved only by the global rank-0 process
        if not is_global_rank_zero:
            return False

        if self.maximise:
            is_best = metric_value > self._best_metric_value
        else:
            is_best = metric_value < self._best_metric_value

        if is_best:
            self._best_metric_value = metric_value
            self._best_metric_epoch = epoch
            self._save_best_metric()

        return is_best

    def should_save_test_outputs(self, is_global_rank_zero: bool = True) -> bool:
        """Determine whether test outputs should be saved.

        :param is_global_rank_zero: Whether this is the global rank-0 process in distributed scenarios.
            Set to `True` (default) if running a single process.
        """
        # This is implemented as a method in case we want to add custom logic in the future
        return is_global_rank_zero


class DeepMILOutputsHandler:
    """Class that manages writing validation and test outputs for DeepMIL models."""

    def __init__(self, outputs_root: Path, n_classes: int, tile_size: int, loading_params: LoadingParams,
                 class_names: Optional[Sequence[str]], primary_val_metric: MetricsKey,
                 maximise: bool, val_plot_options: Collection[PlotOption],
                 test_plot_options: Collection[PlotOption],
                 save_intermediate_outputs: bool = True) -> None:
        """
        :param outputs_root: Root directory where to save all produced outputs.
        :param n_classes: Number of MIL classes (set `n_classes=1` for binary).
        :param tile_size: The size of each tile.
        :param loading_params: Parameters for loading WSI to create plots. This paramter is passed to PlotsHandler.
        :param class_names: List of class names. For binary (`n_classes == 1`), expects `len(class_names) == 2`.
            If `None`, will return `('0', '1', ...)`.
        :param primary_val_metric: Name of the validation metric to track for saving best epoch outputs.
        :param maximise: Whether higher is better for `primary_val_metric`.
        :param val_plot_options: The desired plot options for validation time.
        :param test_plot_options: The desired plot options for test time.
        :param save_intermediate_outputs: Whether to save intermediate outputs (e.g. after each epoch).
        """
        self.outputs_root = outputs_root
        self.n_classes = n_classes
        self.class_names = validate_class_names(class_names, self.n_classes)
        self.outputs_policy = OutputsPolicy(outputs_root=outputs_root,
                                            primary_val_metric=primary_val_metric,
                                            maximise=maximise)
        self.save_intermediate_outputs = save_intermediate_outputs
        self.tiles_selector: Optional[TilesSelector] = None
        self.val_plots_handler = DeepMILPlotsHandler(
            plot_options=val_plot_options,
            tile_size=tile_size,
            class_names=self.class_names,
            stage=ModelKey.VAL,
            loading_params=deepcopy(loading_params),
        )
        self.test_plots_handler = DeepMILPlotsHandler(
            plot_options=test_plot_options,
            tile_size=tile_size,
            class_names=self.class_names,
            stage=ModelKey.TEST,
            loading_params=deepcopy(loading_params),
        )

    @property
    def validation_outputs_dir(self) -> Path:
        return self.outputs_root / VAL_OUTPUTS_SUBDIR

    @property
    def extra_validation_outputs_dir(self) -> Path:
        return self.outputs_root / EXTRA_VAL_OUTPUTS_SUBDIR

    @property
    def previous_validation_outputs_dir(self) -> Path:
        return self.validation_outputs_dir.with_name(PREV_VAL_OUTPUTS_SUBDIR)

    @property
    def test_outputs_dir(self) -> Path:
        return self.outputs_root / TEST_OUTPUTS_SUBDIR

    def set_slides_dataset_for_plots_handlers(self, slides_dataset: Optional[SlidesDataset]) -> None:
        validate_slide_datasets_for_plot_options(self.test_plots_handler.plot_options, slides_dataset)
        validate_slide_datasets_for_plot_options(self.val_plots_handler.plot_options, slides_dataset)
        self.test_plots_handler.slides_dataset = slides_dataset
        self.val_plots_handler.slides_dataset = slides_dataset

    def should_gather_tiles(self, plots_handler: DeepMILPlotsHandler) -> bool:
        return PlotOption.TOP_BOTTOM_TILES in plots_handler.plot_options and self.tiles_selector is not None

    def _save_outputs(self, epoch_results: EpochResultsType, outputs_dir: Path, stage: ModelKey = ModelKey.VAL) -> None:
        """Trigger the rendering and saving of DeepMIL outputs and figures.

        :param epoch_results: Aggregated results from all epoch batches.
        :param outputs_dir: Specific directory into which outputs should be saved (different for validation and test).
        :param stage: The stage of the model (e.g. `ModelKey.VAL` or `ModelKey.TEST`).
        """
        # outputs object consists of a list of dictionaries (of metadata and results, including encoded features)
        # It can be indexed as outputs[batch_idx][batch_key][bag_idx][tile_idx]
        # example of batch_key ResultsKey.SLIDE_ID_COL
        # for batch keys that contains multiple values for slides e.g. ResultsKey.BAG_ATTN_COL
        # outputs[batch_idx][batch_key][bag_idx][tile_idx]
        # contains the tile value
        # TODO: Synchronise this with checkpoint saving (e.g. on_save_checkpoint())
        results = collate_results_on_cpu(epoch_results)
        outputs_dir.mkdir(exist_ok=True, parents=True)
        save_outputs_csv(results, outputs_dir)

        plots_handler = self.val_plots_handler if stage == ModelKey.VAL else self.test_plots_handler
        plots_handler.save_plots(outputs_dir, self.tiles_selector, results)

    def save_validation_outputs(self, epoch_results: EpochResultsType, metrics_dict: Mapping[MetricsKey, Metric],
                                epoch: int, is_global_rank_zero: bool = True, on_extra_val: bool = False) -> None:
        """Render and save validation epoch outputs, according to the configured :py:class:`OutputsPolicy`.

        :param epoch_results: Aggregated results from all epoch batches, as passed to :py:meth:`validation_epoch_end()`.
        :param metrics_dict: Current epoch's validation metrics dictionary from
            :py:class:`~health_cpath.models.deepmil.DeepMILModule`.
        :param is_global_rank_zero: Whether this is the global rank-0 process in distributed scenarios.
            Set to `True` (default) if running a single process.
        :param epoch: Current epoch number.
        :param on_extra_val: Whether this is an extra validation epoch (e.g. after training).
        """
        if self.save_intermediate_outputs or on_extra_val:
            epoch_results = gather_results(epoch_results)

            if self.should_gather_tiles(self.val_plots_handler):
                self.tiles_selector.gather_selected_tiles_across_devices()  # type: ignore

        # Only global rank-0 process should actually render and save the outputs
        if self.save_intermediate_outputs and self.outputs_policy.should_save_validation_outputs(
            metrics_dict, epoch, is_global_rank_zero, on_extra_val
        ):
            # First move existing outputs to a temporary directory, to avoid mixing
            # outputs of different epochs in case writing fails halfway through
            if self.validation_outputs_dir.exists():
                replace_directory(source=self.validation_outputs_dir,
                                  target=self.previous_validation_outputs_dir)

            self._save_outputs(epoch_results, self.validation_outputs_dir, ModelKey.VAL)

            # Writing completed successfully; delete temporary back-up
            if self.previous_validation_outputs_dir.exists():
                shutil.rmtree(self.previous_validation_outputs_dir, ignore_errors=True)
        # We also want to save the plots of the extra validation epoch
        elif on_extra_val and is_global_rank_zero:
            self._save_outputs(epoch_results, self.extra_validation_outputs_dir, ModelKey.VAL)

        # Reset the top and bottom slides heaps
        if self.should_gather_tiles(self.val_plots_handler):
            self.tiles_selector._clear_cached_slides_heaps()  # type: ignore

    def save_test_outputs(self, epoch_results: EpochResultsType, is_global_rank_zero: bool = True) -> None:
        """Render and save test epoch outputs.

        :param epoch_results: Aggregated results from all epoch batches, as passed to :py:meth:`test_epoch_end()`.
        :param metrics_dict: Test metrics dictionary from :py:class:`~health_cpath.models.deepmil.DeepMILModule`.
        :param is_global_rank_zero: Whether this is the global rank-0 process in distributed scenarios.
            Set to `True` (default) if running a single process.
        """
        # All DDP processes must reach this point to allow synchronising epoch results
        gathered_epoch_results = gather_results(epoch_results)
        if self.should_gather_tiles(self.test_plots_handler):
            self.tiles_selector.gather_selected_tiles_across_devices()  # type: ignore

        # Only global rank-0 process should actually render and save the outputs-
        if self.outputs_policy.should_save_test_outputs(is_global_rank_zero):
            self._save_outputs(gathered_epoch_results, self.test_outputs_dir, ModelKey.TEST)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from pathlib import Path
from typing import Mapping, Sequence, Tuple, Union, Callable, Dict

import torch
import numpy as np
import PIL
from PIL import PngImagePlugin
from monai.config.type_definitions import KeysCollection
from monai.transforms import MapTransform, Randomizable, Transform
from monai.utils.enums import WSIPatchKeys
from monai.data.meta_tensor import MetaTensor
from health_azure.logging import elapsed_timer
from health_ml.utils.box_utils import Box
from torchvision.transforms.functional import to_tensor

from health_cpath.models.encoders import TileEncoder
from health_cpath.preprocessing.create_tiles_dataset import get_tile_id
from health_cpath.utils.naming import SlideKey, TileKey

PathOrString = Union[Path, str]


def load_pil_image(image_path: PathOrString) -> PIL.Image.Image:
    """Load a PIL image in RGB format from the given path"""
    with PngImagePlugin.PngImageFile(image_path) as pil_png:
        image = np.asarray(pil_png)
    return image


def load_image_as_tensor(image_path: PathOrString, scale_intensity: bool = True) -> torch.Tensor:
    """Load an image as a tensor from the given path

    :param image_path: path to the image
    :param scale_intensity: if True, use `to_tensor` from torchvision which scales the image pixel intensities to
    [0, 1] by defaul as [C, H, W] tensors. Otherwise, only transpose the image to [C, H, W] format and return it as a
    torch tensor.
    """

    pil_image = load_pil_image(image_path)  # pil_image is in channels last format [H, W, C]
    if scale_intensity:
        return to_tensor(pil_image)  # to_tensor scales the image pixel intensities to [0, 1] as [C, H, W] tensors
    else:
        return torch.from_numpy(pil_image.transpose((2, 0, 1))).contiguous()  # only transpose to [C, H, W]


def load_image_stack_as_tensor(image_paths: Sequence[PathOrString],
                               progress: bool = False,
                               scale_intensity: bool = True) -> torch.Tensor:
    """Load a batch of images of the same size as a tensor from the given paths

    :param image_paths: paths to the images
    :param progress: if True, show a progress bar
    :param scale_intensity: if True, use `to_tensor` from torchvision which scales the image pixel intensities to
    [0, 1] by defaul as [C, H, W] tensors. Otherwise, only transpose the image to [C, H, W] format and return it as a
    torch tensor.
    """

    loading_generator = (load_image_as_tensor(path, scale_intensity) for path in image_paths)
    if progress:
        from tqdm import tqdm
        loading_generator = tqdm(loading_generator, desc="Loading image stack",
                                 total=len(image_paths), leave=False)
    image_tensors = list(loading_generator)
    return torch.stack(image_tensors, dim=0)


def transform_dict_adaptor(function: Callable, k_input: str = None, k_output: str = None) -> Callable:
    """Adapt transformations to work with an input dictionary (rather than a tensor).
       We can't reuse monai.transforms.adaptors because it is only compatible with transformations that accept
       a dict as input.

    :param function: a transformation function
    :param k_input: key of the input dictionary that contains the object
        to which function should be applied
    :param k_output: key of the input dictionary where to place the function output. If None the ouput of
        the transformation is returned

    :return: adapted transformation
    """
    def _inner(ditems: dict) -> Dict:
        if k_input is None:
            dinputs = ditems
        else:
            dinputs = ditems[k_input]
        ret = function(dinputs)
        if k_output is None:
            ditems = ret
        else:
            if isinstance(ret, type(ditems[k_output])):
                ditems[k_output] = ret
            else:
                raise ValueError("The transformation is not expect to change the type."
                                 "Check input and output are used correctly ")
        return ditems
    return _inner


class LoadTiled(MapTransform):
    """Dictionary transform to load an individual image tile as a tensor from an input path"""

    def __init__(self, keys: KeysCollection, allow_missing_keys: bool = False) -> None:
        """
        :param keys: Key(s) for the image path(s) in the input dictionary.
        :param allow_missing_keys: If `False` (default), raises an exception when an input
        dictionary is missing any of the specified keys.
        """
        super().__init__(keys, allow_missing_keys)

    def __call__(self, data: Mapping) -> Mapping:
        out_data = dict(data)  # create shallow copy
        for key in self.key_iterator(out_data):
            out_data[key] = load_image_as_tensor(data[key])
        return out_data


class LoadTilesBatchd(MapTransform):
    """Dictionary transform to load a batch of image tiles as a tensor from a list of input paths"""

    # Cannot reuse MONAI readers because they support stacking only images with no channels
    def __init__(self, keys: KeysCollection, allow_missing_keys: bool = False,
                 progress: bool = False, scale_intensity: bool = True) -> None:
        """
        :param keys: Key(s) for the image path(s) in the input dictionary.
        :param allow_missing_keys: If `False` (default), raises an exception when an input
        dictionary is missing any of the specified keys.
        :param progress: Whether to display a tqdm progress bar.
        :param scale_intensity: if True, use `to_tensor` from torchvision which scales the image pixel intensities to
        [0, 1] by defaul as [C, H, W] tensors. Otherwise, only transpose the image to [C, H, W] format and return it as
        a torch tensor.
        """

        super().__init__(keys, allow_missing_keys)
        self.progress = progress
        self.scale_intensity = scale_intensity

    def __call__(self, data: Mapping) -> Mapping:
        out_data = dict(data)  # create shallow copy
        for key in self.key_iterator(out_data):
            out_data[key] = load_image_stack_as_tensor(
                data[key], progress=self.progress, scale_intensity=self.scale_intensity
            )
        return out_data


class EncodeTilesBatchd(MapTransform):
    """Dictionary transform to extract features from a batch tensor of image tiles"""

    def __init__(self,
                 keys: KeysCollection,
                 encoder: TileEncoder,
                 allow_missing_keys: bool = False,
                 chunk_size: int = 0) -> None:
        """
        :param keys: Key(s) for the image tensor(s) in the input dictionary.
        :param encoder: The tile encoder to use for feature extraction.
        :param allow_missing_keys: If `False` (default), raises an exception when an input
        dictionary is missing any of the specified keys.
        :param chunk_size: if > 0, extracts features in chunks of size chunk_size.
        """
        super().__init__(keys, allow_missing_keys)
        self.encoder = encoder
        self.chunk_size = chunk_size

    @torch.no_grad()
    def _encode_tiles(self, images: torch.Tensor) -> torch.Tensor:
        device = next(self.encoder.parameters()).device
        if self.chunk_size > 0:
            embeddings = []
            chunks = torch.split(images, self.chunk_size)
            # TODO parallelize encoding - keep metadata and images aligned
            for chunk in chunks:
                chunk_embeddings = self._encode_images(chunk, device)
                embeddings.append(chunk_embeddings)
            return torch.cat(embeddings)
        else:
            return self._encode_images(images, device)

    def _encode_images(self, images: torch.Tensor, device: torch.device) -> torch.Tensor:
        images = images.to(device)
        embeddings = self.encoder(images)
        del images
        torch.cuda.empty_cache()
        return embeddings

    def __call__(self, data: Mapping) -> Mapping:
        out_data = dict(data)  # create shallow copy
        for key in self.key_iterator(out_data):
            out_data[key] = self._encode_tiles(data[key])
        return out_data


def take_indices(data: Sequence, indices: np.ndarray) -> Sequence:
    if isinstance(data, (np.ndarray, torch.Tensor)):
        return data[indices]  # type: ignore
    elif isinstance(data, Sequence):
        return [data[i] for i in indices]
    else:
        raise ValueError(f"Data of type {type(data)} is not indexable")


class Subsampled(MapTransform, Randomizable):
    """Dictionary transform to randomly subsample the data down to a fixed maximum length"""

    def __init__(self, keys: KeysCollection, max_size: int,
                 allow_missing_keys: bool = False) -> None:
        """
        :param keys: Key(s) for all batch elements that must be subsampled.
        :param max_size: Each specified array, tensor, or sequence will be subsampled uniformly at
        random down to `max_size` along their first dimension. If shorter, the elements are merely
        shuffled.
        :param allow_missing_keys: If `False` (default), raises an exception when an input
        dictionary is missing any of the specified keys.
        """
        super().__init__(keys, allow_missing_keys=allow_missing_keys)
        self.max_size = max_size
        self._indices: np.ndarray

    def randomize(self, total_size: int) -> None:
        subsample_size = min(self.max_size, total_size)
        self._indices = self.R.choice(total_size, size=subsample_size, replace=False)

    def __call__(self, data: Mapping) -> Mapping:
        out_data = dict(data)  # create shallow copy
        size = len(data[self.keys[0]])
        self.randomize(size)
        for key in self.key_iterator(out_data):
            out_data[key] = take_indices(data[key], self._indices)
        return out_data


class ExtractCoordinatesd(MapTransform):
    """Extract the coordinates of the tiles returned as meta data by monai transforms to hi-ml-cpath format where
    the coordinates are represented as TileKey.TILE_LEFT, TileKey.TILE_TOP, TileKey.TILE_RIGHT, TileKey.TILE_BOTTOM."""

    def __init__(self, image_key: str, tile_size: int) -> None:
        self.tile_size = tile_size
        self.image_key = image_key

    def extract_coordinates(self, data: Dict) -> Tuple[np.ndarray, np.ndarray]:
        """Extract the coordinates of the tiles from the metadata."""
        assert isinstance(data[self.image_key], MetaTensor), f"Expected MetaTensor, got {type(data[self.image_key])}"
        ys, xs = data[self.image_key].meta[WSIPatchKeys.LOCATION]
        return ys, xs

    def extract_scale_factor(self, data: Dict) -> int:
        """Extract the scale factor of the tiles from the metadata to rescale the coordinates to highest resolution."""
        return int(data.get(SlideKey.SCALE, 1))

    def extract_offset(self, data: Dict) -> Tuple[int, int]:
        """Extract the offset of the tiles from the metadata to translate to (0, 0) origin."""
        return data.get(SlideKey.ORIGIN, (0, 0))

    def set_coordinates(self, data: Dict, xs: np.ndarray, ys: np.ndarray) -> None:
        """Set the coordinates of the tiles in the metadata."""
        # Extract the scale factor and offset to rescale the coordinates to highest resolution
        scale_factor = self.extract_scale_factor(data)
        offset_y, offset_x = self.extract_offset(data)
        # We set the coordinates of the tiles as top left and bottom right coordinates
        data[TileKey.TILE_LEFT] = torch.tensor((xs * scale_factor + offset_x))
        data[TileKey.TILE_TOP] = torch.tensor((ys * scale_factor + offset_y))
        data[TileKey.TILE_RIGHT] = data[TileKey.TILE_LEFT] + self.tile_size * scale_factor
        data[TileKey.TILE_BOTTOM] = data[TileKey.TILE_TOP] + self.tile_size * scale_factor

    def set_tile_and_slide_ids(self, data: Dict, xs: np.ndarray, ys: np.ndarray) -> None:
        """Set the tile and slide id in the metadata."""
        h, w = self.tile_size, self.tile_size
        bag_size = data[SlideKey.IMAGE].meta[WSIPatchKeys.COUNT]
        data[TileKey.TILE_ID] = [get_tile_id(data[SlideKey.SLIDE_ID], Box(x=x, y=y, w=w, h=h)) for x, y in zip(xs, ys)]
        data[SlideKey.SLIDE_ID] = [data[SlideKey.SLIDE_ID]] * bag_size

    def convert_tiles_and_label_to_tensors(self, data: Dict) -> None:
        """Convert the tiles and label to tensors."""
        data[SlideKey.IMAGE] = data[SlideKey.IMAGE].as_tensor()
        data[SlideKey.LABEL] = torch.tensor(data[SlideKey.LABEL])

    def __call__(self, data: Mapping) -> Mapping:
        out_data = dict(data)
        ys, xs = self.extract_coordinates(out_data)
        self.set_coordinates(out_data, xs=xs, ys=ys)
        self.set_tile_and_slide_ids(out_data, xs=xs, ys=ys)
        self.convert_tiles_and_label_to_tensors(out_data)
        return out_data


class MetaTensorToTensord(MapTransform):
    """Converts a MetaTensor to a Tensor."""

    def __call__(self, data: Mapping) -> Mapping:
        out_data = dict(data)
        for key in self.key_iterator(out_data):
            assert isinstance(out_data[key], MetaTensor), f"Expected MetaTensor, got {type(out_data[key])}"
            out_data[key] = out_data[key].as_tensor()
        return out_data


class TimerWrapper(Transform):
    """Transform that measures the time it takes to execute the transform. Useful for debugging. This can be used by
    wrapping the transform in a TimerWrapperd transform.
    Example:
        transform = Compose([TimerWrapperd(LoadImaged(keys="image")), TimerWrapperd(MetaTensorToTensord(keys="image"))])
    """

    def __init__(self, transform: Callable) -> None:
        self.transform = transform

    def __call__(self, data: Mapping) -> Mapping:
        message = f"{self.transform.__class__.__name__}, Slide {data[SlideKey.SLIDE_ID]}"
        with elapsed_timer(message):
            out_data = self.transform(data)
        return out_data



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

from pathlib import Path
from typing import Optional, Union

import pandas as pd

from health_cpath.datasets.base_dataset import DEFAULT_LABEL_COLUMN, SlidesDataset


class TcgaPradDataset(SlidesDataset):
    """Dataset class for loading TCGA-PRAD slides.

    Iterating over this dataset returns a dictionary containing:
    - `'slide_id'` (str)
    - `'case_id'` (str)
    - `'image_path'` (str): absolute slide image path
    - `'label'` (int, 0 or 1): label for predicting positive or negative
    """
    IMAGE_COLUMN: str = 'image_path'

    DEFAULT_CSV_FILENAME: str = "dataset.csv"

    def __init__(self, root: Union[str, Path],
                 dataset_csv: Optional[Union[str, Path]] = None,
                 dataset_df: Optional[pd.DataFrame] = None,
                 label_column: str = DEFAULT_LABEL_COLUMN) -> None:
        """
        :param root: Root directory of the dataset.
        :param dataset_csv: Full path to a dataset CSV file. If omitted, the CSV will be read from
        `"{root}/{DEFAULT_CSV_FILENAME}"`.
        :param dataset_df: A potentially pre-processed dataframe in the same format as would be read
        from the dataset CSV file, e.g. after some filtering. If given, overrides `dataset_csv`.
        """
        super().__init__(root, dataset_csv, dataset_df, validate_columns=False, label_column=label_column)
        # Example of how to define a custom label column from existing columns:
        self.dataset_df[self.label_column] = (self.dataset_df['label1']
                                              | self.dataset_df['label2']).astype(int)  # noqa: W503
        self.validate_columns()



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from pathlib import Path
import sys
import time
from typing import Any, Optional
from azureml.core import Workspace

himl_histo_root_dir = Path(__file__).parent.parent.parent
himl_root = himl_histo_root_dir.parent.parent
himl_azure_package_root = himl_root / "hi-ml-azure" / "src"
sys.path.insert(0, str(himl_azure_package_root))

from health_azure import DatasetConfig  # noqa: E402
from health_azure.utils import get_workspace  # noqa: E402


def mount_dataset(dataset_id: str, tmp_root: str = "/tmp/datasets", aml_workspace: Optional[Workspace] = None) -> Any:
    ws = get_workspace(aml_workspace)
    target_folder = "/".join([tmp_root, dataset_id])
    dataset = DatasetConfig(name=dataset_id, target_folder=target_folder, use_mounting=True)
    _, mount_ctx = dataset.to_input_dataset_local(strictly_aml_v1=True, workspace=ws)
    assert mount_ctx is not None  # for mypy
    mount_ctx.start()
    return mount_ctx


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    # Run this script as "python mount_azure_dataset.py --dataset_id TCGA-CRCk"
    parser.add_argument('--dataset_id', type=str,
                        help='Name of the Azure dataset e.g. PANDA or TCGA-CRCk')
    args = parser.parse_args()
    # It is essential that the mount context is returned from the mounting function and referenced here.
    # If not, mounting will be stopped, and the files are no longer available.
    _ = mount_dataset(args.dataset_id)
    print("The mounted dataset will only be available while this script is running. Press Ctrl-C to terminate it.`")
    while True:
        time.sleep(60)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------


from typing import Any, Dict, Callable, Union
from torch import optim
from monai.transforms import Compose, ScaleIntensityRanged, RandRotate90d, RandFlipd
from health_cpath.configs.run_ids import innereye_ssl_checkpoint_binary
from health_azure.utils import create_from_matching_params
from health_cpath.models.transforms import MetaTensorToTensord
from health_cpath.preprocessing.loading import LoadingParams
from health_cpath.utils.wsi_utils import TilingParams
from health_ml.networks.layers.attention_layers import (
    TransformerPooling,
    TransformerPoolingBenchmark
)
from health_ml.utils.checkpoint_utils import CheckpointParser
from health_ml.deep_learning_config import OptimizerParams
from health_cpath.datasets.panda_dataset import PandaDataset
from health_cpath.datamodules.panda_module_benchmark import PandaSlidesDataModuleBenchmark
from health_cpath.models.encoders import (
    HistoSSLEncoder,
    ImageNetSimCLREncoder,
    Resnet50_NoPreproc,
    SSLEncoder,
)
from health_cpath.configs.classification.DeepSMILEPanda import DeepSMILESlidesPanda
from health_cpath.models.deepmil import DeepMILModule
from health_cpath.utils.deepmil_utils import ClassifierParams, EncoderParams, PoolingParams
from health_cpath.utils.naming import MetricsKey, ModelKey, SlideKey


class PandaSlidesDeepMILModuleBenchmark(DeepMILModule):
    """
    Myronenko et al. 2021 uses a cosine LR scheduler which needs to be defined in the PL module
    Hence, inherited `PandaSlidesDeepMILModuleBenchmark` from `DeepMILModule`
    """

    def __init__(self, n_epochs: int, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.save_hyperparameters()
        self.n_epochs = n_epochs

    def configure_optimizers(self) -> Dict[str, Any]:           # type: ignore
        optimizer = optim.AdamW(self.parameters(), lr=self.optimizer_params.l_rate,
                                weight_decay=self.optimizer_params.weight_decay)
        scheduler = optim.lr_scheduler.CosineAnnealingLR(optimizer=optimizer, T_max=self.n_epochs, eta_min=0)
        return {"optimizer": optimizer, "lr_scheduler": scheduler}


class DeepSMILESlidesPandaBenchmark(DeepSMILESlidesPanda):
    """
    Configuration for PANDA experiments from Myronenko et al. 2021:
    (https://link.springer.com/chapter/10.1007/978-3-030-87237-3_32)
    `tune_encoder` sets the fine-tuning mode of the encoder. For fine-tuning, batch_size = 2 runs on 8 GPUs
    with ~ 6:24 min/epoch (train) and ~ 00:50 min/epoch (validation).
    """

    def __init__(self, **kwargs: Any) -> None:
        default_kwargs = dict(
            pool_type=TransformerPoolingBenchmark.__name__,
            num_transformer_pool_layers=4,
            num_transformer_pool_heads=8,
            pool_hidden_dim=2048,
            encoding_chunk_size=60,
            max_bag_size=56,
            batch_size=8,  # effective batch size = batch_size * num_GPUs
            batch_size_inf=8,
            max_epochs=50,
            l_rate=3e-4,
            weight_decay=0,
            primary_val_metric=MetricsKey.ACC)
        default_kwargs.update(kwargs)
        super().__init__(**default_kwargs)

    def setup(self) -> None:
        # Params specific to transformer pooling
        if self.pool_type in [TransformerPoolingBenchmark.__name__, TransformerPooling.__name__]:
            self.l_rate = 3e-5
            self.weight_decay = 0.1
        # Params specific to fine-tuning
        if self.tune_encoder:
            self.batch_size = 2
            self.batch_size_inf = 2
        super().setup()

    def get_transforms_dict(self, image_key: str) -> Dict[ModelKey, Union[Callable, None]]:
        # Use same transforms as demonstrated in
        # https://github.com/Project-MONAI/tutorials/blob/master/pathology/multiple_instance_learning/panda_mil_train_evaluate_pytorch_gpu.py
        transform_train = Compose([
            RandFlipd(keys=image_key, spatial_axis=0, prob=0.5),
            RandFlipd(keys=image_key, spatial_axis=1, prob=0.5),
            RandRotate90d(keys=image_key, prob=0.5),
            ScaleIntensityRanged(keys=image_key, a_min=0.0, a_max=255.0),
            MetaTensorToTensord(keys=image_key),  # rotate transforms add some metadata to affine matrix
        ])
        transform_inf = Compose([
            ScaleIntensityRanged(keys=image_key, a_min=0.0, a_max=255.0),
        ])
        return {ModelKey.TRAIN: transform_train, ModelKey.VAL: transform_inf, ModelKey.TEST: transform_inf}

    def get_data_module(self) -> PandaSlidesDataModuleBenchmark:  # type: ignore
        # Myronenko et al. 2021 uses 80-20 cross-val split and no hold-out test set
        # Hence, inherited `PandaSlidesDataModuleBenchmark` from `SlidesDataModule`
        return PandaSlidesDataModuleBenchmark(
            root_path=self.local_datasets[0],
            batch_size=self.batch_size,
            batch_size_inf=self.batch_size_inf,
            max_bag_size=self.max_bag_size,
            max_bag_size_inf=self.max_bag_size_inf,
            tiling_params=create_from_matching_params(self, TilingParams),
            loading_params=create_from_matching_params(self, LoadingParams),
            seed=self.get_effective_random_seed(),
            transforms_dict=self.get_transforms_dict(PandaDataset.IMAGE_COLUMN),
            crossval_count=self.crossval_count,
            crossval_index=self.crossval_index,
            dataloader_kwargs=self.get_dataloader_kwargs(),
            pl_replace_sampler_ddp=self.pl_replace_sampler_ddp,
        )

    def create_model(self) -> DeepMILModule:
        self.data_module = self.get_data_module()
        outputs_handler = self.get_outputs_handler()
        deepmil_module = PandaSlidesDeepMILModuleBenchmark(
            n_epochs=self.max_epochs,
            label_column=SlideKey.LABEL,
            n_classes=self.data_module.train_dataset.n_classes,
            class_names=self.class_names,
            class_weights=self.data_module.class_weights,
            outputs_folder=self.outputs_folder,
            encoder_params=create_from_matching_params(self, EncoderParams),
            pooling_params=create_from_matching_params(self, PoolingParams),
            classifier_params=create_from_matching_params(self, ClassifierParams),
            optimizer_params=create_from_matching_params(self, OptimizerParams),
            outputs_handler=outputs_handler,
            analyse_loss=self.analyse_loss,
        )
        outputs_handler.set_slides_dataset_for_plots_handlers(self.get_slides_dataset())
        return deepmil_module


class SlidesPandaImageNetMILBenchmark(DeepSMILESlidesPandaBenchmark):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=Resnet50_NoPreproc.__name__, **kwargs)


class SlidesPandaImageNetSimCLRMILBenchmark(DeepSMILESlidesPandaBenchmark):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=ImageNetSimCLREncoder.__name__, **kwargs)


class SlidesPandaSSLMILBenchmark(DeepSMILESlidesPandaBenchmark):
    def __init__(self, **kwargs: Any) -> None:
        # If no SSL checkpoint is provided, use the default one
        self.ssl_checkpoint = self.ssl_checkpoint or CheckpointParser(innereye_ssl_checkpoint_binary)
        super().__init__(encoder_type=SSLEncoder.__name__, **kwargs)


class SlidesPandaHistoSSLMILBenchmark(DeepSMILESlidesPandaBenchmark):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=HistoSSLEncoder.__name__, **kwargs)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

# These tiling implementations are adapted from PANDA Kaggle solutions, for example:
# https://github.com/kentaroy47/Kaggle-PANDA-1st-place-solution/blob/master/src/data_process/a00_save_tiles.py
from typing import Any, Optional, Tuple

import numpy as np


def get_1d_padding(length: int, tile_size: int) -> Tuple[int, int]:
    """Computes symmetric padding for `length` to be divisible by `tile_size`."""
    pad = (tile_size - length % tile_size) % tile_size
    return (pad // 2, pad - pad // 2)


def pad_for_tiling_2d(array: np.ndarray, tile_size: int, channels_first: Optional[bool] = True,
                      **pad_kwargs: Any) -> Tuple[np.ndarray, np.ndarray]:
    """Symmetrically pads a 2D `array` such that both dimensions are divisible by `tile_size`.

    :param array: 2D image array.
    :param tile_size: Width/height of each tile in pixels.
    :param channels_first: Whether `array` is in CHW (`True`, default) or HWC (`False`) layout.
    :param pad_kwargs: Keyword arguments to be passed to `np.pad()` (e.g. `constant_values=0`).
    :return: A tuple containing:
        - `padded_array`: Resulting array, in the same CHW/HWC layout as the input.
        - `offset`: XY offset introduced by the padding. Add this to coordinates relative to the
        original array to obtain indices for the padded array.
    """
    height, width = array.shape[1:] if channels_first else array.shape[:-1]
    padding_h = get_1d_padding(height, tile_size)
    padding_w = get_1d_padding(width, tile_size)
    padding = [padding_h, padding_w]
    channels_axis = 0 if channels_first else 2
    padding.insert(channels_axis, (0, 0))  # zero padding on channels axis
    padded_array = np.pad(array, padding, **pad_kwargs)
    offset = (padding_w[0], padding_h[0])
    return padded_array, np.array(offset)


def tile_array_2d(array: np.ndarray, tile_size: int, channels_first: Optional[bool] = True,
                  **pad_kwargs: Any) -> Tuple[np.ndarray, np.ndarray]:
    """Split an image array into square non-overlapping tiles.

    The array will be padded symmetrically if its dimensions are not exact multiples of `tile_size`.

    :param array: Image array.
    :param tile_size: Width/height of each tile in pixels.
    :param pad_kwargs: Keyword arguments to be passed to `np.pad()` (e.g. `constant_values=0`).
    :param channels_first: Whether `array` is in CHW (`True`, default) or HWC (`False`) layout.
    :return: A tuple containing:
        - `tiles`: A batch of tiles in NCHW layout.
        - `coords`: XY coordinates of each tile, in the same order.
    """
    padded_array, (offset_w, offset_h) = pad_for_tiling_2d(array, tile_size, channels_first, **pad_kwargs)
    if channels_first:
        channels, height, width = padded_array.shape
    else:
        height, width, channels = padded_array.shape
    n_tiles_h = height // tile_size
    n_tiles_w = width // tile_size

    if channels_first:
        intermediate_shape = (channels, n_tiles_h, tile_size, n_tiles_w, tile_size)
        axis_order = (1, 3, 0, 2, 4)  # (n_tiles_h, n_tiles_w, channels, tile_size, tile_size)
        output_shape = (n_tiles_h * n_tiles_w, channels, tile_size, tile_size)
    else:
        intermediate_shape = (n_tiles_h, tile_size, n_tiles_w, tile_size, channels)
        axis_order = (0, 2, 1, 3, 4)  # (n_tiles_h, n_tiles_w, tile_size, tile_size, channels)
        output_shape = (n_tiles_h * n_tiles_w, tile_size, tile_size, channels)

    tiles = padded_array.reshape(intermediate_shape)  # Split width and height axes
    tiles = tiles.transpose(axis_order)
    tiles = tiles.reshape(output_shape)  # Flatten tile batch dimension

    # Compute top-left coordinates of every tile, relative to the original array's origin
    coords_h = tile_size * np.arange(n_tiles_h) - offset_h
    coords_w = tile_size * np.arange(n_tiles_w) - offset_w
    # Shape: (n_tiles_h * n_tiles_w, 2)
    coords = np.stack(np.meshgrid(coords_w, coords_h), axis=-1).reshape(-1, 2)

    return tiles, coords


def assemble_tiles_2d(tiles: np.ndarray, coords: np.ndarray, fill_value: Optional[float] = np.nan,
                      channels_first: Optional[bool] = True) -> Tuple[np.ndarray, np.ndarray]:
    """Assembles a 2D array from sequences of tiles and coordinates.

    :param tiles: Stack of tiles with batch dimension first.
    :param coords: XY tile coordinates, assumed to be spaced by multiples of `tile_size` (shape: [N, 2]).
    :param tile_size: Size of each tile; must be >0.
    :param fill_value: Value to assign to empty elements (default: `NaN`).
    :param channels_first: Whether each tile is in CHW (`True`, default) or HWC (`False`) layout.
    :return: A tuple containing:
        - `array`: The reassembled 2D array with the smallest dimensions to contain all given tiles.
        - `offset`: The lowest XY coordinates.
        - `offset`: XY offset introduced by the assembly. Add this to tile coordinates to obtain
        indices for the assembled array.
    """
    if coords.shape[0] != tiles.shape[0]:
        raise ValueError(f"Tile coordinates and values must have the same length, "
                         f"got {coords.shape[0]} and {tiles.shape[0]}")

    if channels_first:
        n_tiles, channels, tile_size, _ = tiles.shape
    else:
        n_tiles, tile_size, _, channels = tiles.shape
    tile_xs, tile_ys = coords.T

    x_min, x_max = min(tile_xs), max(tile_xs + tile_size)
    y_min, y_max = min(tile_ys), max(tile_ys + tile_size)
    width = x_max - x_min
    height = y_max - y_min
    output_shape = (channels, height, width) if channels_first else (height, width, channels)
    array = np.full(output_shape, fill_value)

    offset = np.array([-x_min, -y_min])
    for idx in range(n_tiles):
        row = coords[idx, 1] + offset[1]
        col = coords[idx, 0] + offset[0]
        if channels_first:
            array[:, row:row + tile_size, col:col + tile_size] = tiles[idx]
        else:
            array[row:row + tile_size, col:col + tile_size, :] = tiles[idx]

    return array, offset



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

import os
import torch
import param
import logging
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.cm as cm

from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import Callback

from health_cpath.models.deepmil import DeepMILModule
from health_cpath.utils.naming import ModelKey, ResultsKey
from health_cpath.utils.output_utils import BatchResultsType

LossCacheDictType = Dict[Union[ResultsKey, str], List]
LossDictType = Dict[str, List]
AnomalyDictType = Dict[ModelKey, List[str]]


class LossCallbackParams(param.Parameterized):
    """Parameters class to group all attributes for loss values analysis callback"""

    analyse_loss: bool = param.Boolean(
        False,
        doc="If True, will use `LossValuesAnalysisCallback` to cache loss values per slide/epoch for further analysis."
        "See `LossValuesAnalysisCallback` for more details.",
    )
    loss_analysis_patience: int = param.Integer(
        0,
        bounds=(0, None),
        doc="Number of epochs to wait before starting loss values per slide analysis. Default: 0, It will start"
        "caching loss values per epoch immediately. Use loss_analysis_patience=n>0 to wait for a few epochs "
        "before starting the analysis.",
    )
    loss_analysis_epochs_interval: int = param.Integer(
        1,
        bounds=(1, None),
        doc="Epochs interval to save loss values. Default: 1, It will save loss values every epoch. Use "
        "loss_analysis_epochs_interval=n>1 to save loss values every n epochs.",
    )
    num_slides_scatter: int = param.Integer(
        20,
        bounds=(1, None),
        doc="Number of slides to plot in the scatter plot. Default: 10, It will plot a scatter of the 10 slides "
        "with highest/lowest loss values across epochs.",
    )
    num_slides_heatmap: int = param.Integer(
        20,
        bounds=(1, None),
        doc="Number of slides to plot in the heatmap plot. Default: 20, It will plot the loss values for the 20 slides "
        "with highest/lowest loss values.",
    )
    save_tile_ids: bool = param.Boolean(
        True,
        doc="If True, will save the tile ids for each bag in the loss cache. Default: True. If False, will only save "
        "the slide ids and their loss values.",
    )
    log_exceptions: bool = param.Boolean(
        True,
        doc="If True, will log exceptions raised during loss values analysis. Default: True. If False, will raise the "
        "intercepted exceptions.",
    )


class LossAnalysisCallback(Callback):
    """Callback to analyse loss values per slide across epochs. It saves the loss values per slide in a csv file every n
    epochs and plots the evolution of the loss values per slide in a heatmap as well as the slides with the
    highest/lowest loss values per epoch in a scatter plot."""

    TILES_JOIN_TOKEN = "$"
    X_LABEL, Y_LABEL = "Epoch", "Slide ids"
    TOP, BOTTOM = "top", "bottom"
    HIGHEST, LOWEST = "highest", "lowest"

    def __init__(
        self,
        outputs_folder: Path,
        max_epochs: int = 30,
        patience: int = 0,
        epochs_interval: int = 1,
        num_slides_scatter: int = 10,
        num_slides_heatmap: int = 20,
        save_tile_ids: bool = False,
        log_exceptions: bool = True,
        create_outputs_folders: bool = True,
    ) -> None:
        """

        :param outputs_folder: Path to the folder where the outputs will be saved.
        :param patience: Number of epochs to wait before starting to cache loss values, defaults to 0.
        :param epochs_interval: Epochs interval to save loss values, defaults to 1.
        :param max_epochs: Maximum number of epochs to train, defaults to 30.
        :param num_slides_scatter: Number of slides to plot in the scatter plot, defaults to 10.
        :param num_slides_heatmap: Number of slides to plot in the heatmap, defaults to 20.
        :param save_tile_ids: If True, will save the tile ids of the tiles in the bag in the loss cache, defaults to
        False. This is useful to analyse the tiles that are contributing to the loss value of a slide.
        :param log_exceptions: If True, will log exceptions raised during loss values analysis, defaults to True. If
        False will raise the intercepted exceptions.
        :param create_outputs_folders: If True, will create the output folders if they don't exist, defaults to True.
        """

        self.patience = patience
        self.epochs_interval = epochs_interval
        self.max_epochs = max_epochs
        self.num_slides_scatter = num_slides_scatter
        self.num_slides_heatmap = num_slides_heatmap
        self.save_tile_ids = save_tile_ids
        self.log_exceptions = log_exceptions

        self.outputs_folder = outputs_folder / "loss_analysis_callback"
        if create_outputs_folders:
            self.create_outputs_folders()

        self.loss_cache: Dict[ModelKey, LossCacheDictType] = {
            stage: self.get_empty_loss_cache() for stage in [ModelKey.TRAIN, ModelKey.VAL]
        }
        self.epochs_range = list(range(self.patience, self.max_epochs, self.epochs_interval))

        self.nan_slides: AnomalyDictType = {stage: [] for stage in [ModelKey.TRAIN, ModelKey.VAL]}
        self.anomaly_slides: AnomalyDictType = {stage: [] for stage in [ModelKey.TRAIN, ModelKey.VAL]}

    def get_cache_folder(self, stage: ModelKey) -> Path:
        return self.outputs_folder / f"{stage}/loss_cache"

    def get_scatter_folder(self, stage: ModelKey) -> Path:
        return self.outputs_folder / f"{stage}/loss_scatter"

    def get_heatmap_folder(self, stage: ModelKey) -> Path:
        return self.outputs_folder / f"{stage}/loss_heatmap"

    def get_stats_folder(self, stage: ModelKey) -> Path:
        return self.outputs_folder / f"{stage}/loss_stats"

    def get_anomalies_folder(self, stage: ModelKey) -> Path:
        return self.outputs_folder / f"{stage}/loss_anomalies"

    def create_outputs_folders(self) -> None:
        """Creates the output folders if they don't exist."""
        folders = [
            self.get_cache_folder,
            self.get_scatter_folder,
            self.get_heatmap_folder,
            self.get_stats_folder,
            self.get_anomalies_folder,
        ]
        stages = [ModelKey.TRAIN, ModelKey.VAL]
        for folder in folders:
            for stage in stages:
                os.makedirs(folder(stage), exist_ok=True)

    def get_empty_loss_cache(self) -> LossCacheDictType:
        """Returns an empty loss cache dictionary for keys: slide_id, loss, entropy and tile_ids if save_tile_ids."""
        keys = [ResultsKey.SLIDE_ID, ResultsKey.LOSS, ResultsKey.ENTROPY]
        if self.save_tile_ids:
            keys.append(ResultsKey.TILE_ID)
        return {key: [] for key in keys}

    def _format_epoch(self, epoch: int) -> str:
        """Formats the epoch number to a string with 3 digits."""
        return str(epoch).zfill(len(str(self.max_epochs)))

    def get_loss_cache_file(self, epoch: int, stage: ModelKey) -> Path:
        return self.get_cache_folder(stage) / f"epoch_{self._format_epoch(epoch)}.csv"

    def get_all_epochs_loss_cache_file(self, stage: ModelKey) -> Path:
        return self.get_cache_folder(stage) / f"all_epochs_{stage}.csv"

    def get_loss_stats_file(self, stage: ModelKey) -> Path:
        return self.get_stats_folder(stage) / f"loss_stats_{stage}.csv"

    def get_loss_ranks_file(self, stage: ModelKey) -> Path:
        return self.get_stats_folder(stage) / f"loss_ranks_{stage}.csv"

    def get_loss_ranks_stats_file(self, stage: ModelKey) -> Path:
        return self.get_stats_folder(stage) / f"loss_ranks_stats_{stage}.csv"

    def get_nan_slides_file(self, stage: ModelKey) -> Path:
        return self.get_anomalies_folder(stage) / f"nan_slides_{stage}.txt"

    def get_anomaly_slides_file(self, stage: ModelKey) -> Path:
        return self.get_anomalies_folder(stage) / f"anomaly_slides_{stage}.txt"

    def get_scatter_plot_file(self, order: str, stage: ModelKey) -> Path:
        return self.get_scatter_folder(stage) / f"slides_with_{order}_loss_values_{stage}.png"

    def get_heatmap_plot_file(self, epoch: int, order: str, stage: ModelKey) -> Path:
        return self.get_heatmap_folder(stage) / f"epoch_{self._format_epoch(epoch)}_{order}_slides_{stage}.png"

    def read_loss_cache(self, epoch: int, stage: ModelKey, idx_col: Optional[ResultsKey] = None) -> pd.DataFrame:
        columns = [ResultsKey.SLIDE_ID, ResultsKey.LOSS, ResultsKey.ENTROPY]
        return pd.read_csv(self.get_loss_cache_file(epoch, stage), index_col=idx_col, usecols=columns)

    def should_cache_loss_values(self, current_epoch: int) -> bool:
        """Returns True if the current epoch is a multiple of the epochs_interval."""
        if current_epoch >= self.max_epochs:
            return False  # Don't cache loss values for the extra validation epoch
        current_epoch = current_epoch + 1
        first_time = (current_epoch - self.patience) == 1
        return first_time or (
            current_epoch > self.patience and (current_epoch - self.patience) % self.epochs_interval == 0
        )

    def merge_loss_caches(self, loss_caches: List[LossCacheDictType]) -> LossCacheDictType:
        """Merges the loss caches from all the workers into a single loss cache"""
        loss_cache = self.get_empty_loss_cache()
        for loss_cache_per_device in loss_caches:
            for key in loss_cache.keys():
                loss_cache[key].extend(loss_cache_per_device[key])
        return loss_cache

    def gather_loss_cache(self, rank: int, stage: ModelKey) -> None:
        """Gathers the loss cache from all the workers"""
        if torch.distributed.is_initialized():
            world_size = torch.distributed.get_world_size()
            if world_size > 1:
                loss_caches = [None] * world_size
                torch.distributed.all_gather_object(loss_caches, self.loss_cache[stage])
                if rank == 0:
                    self.loss_cache[stage] = self.merge_loss_caches(loss_caches)  # type: ignore

    def save_loss_cache(self, current_epoch: int, stage: ModelKey) -> None:
        """Saves the loss cache to a csv file"""
        loss_cache_df = pd.DataFrame(self.loss_cache[stage])
        # Some slides may be appear multiple times in the loss cache in DDP mode. The Distributed Sampler duplicate
        # some slides to even out the number of samples per device, so we only keep the first occurrence.
        loss_cache_df.drop_duplicates(subset=ResultsKey.SLIDE_ID, inplace=True, keep="first")
        loss_cache_df = loss_cache_df.sort_values(by=ResultsKey.LOSS, ascending=False)
        loss_cache_df.to_csv(self.get_loss_cache_file(current_epoch, stage), index=False)

    def _select_values_for_epoch(
        self,
        keys: List[ResultsKey],
        epoch: int,
        stage: ModelKey,
        high: Optional[bool] = None,
        num_values: Optional[int] = None
    ) -> List[np.ndarray]:
        """Selects the values corresponding to keys from a dataframe for the given epoch and stage.

        :param keys: The keys to select.
        :param epoch: The epoch to select.
        :param stage: The model's stage e.g. train, val, test.
        :param high: If True, selects the highest values, if False, selects the lowest values, if None, selects all
            values.
        :param num_values: The number of values to select.
        """
        loss_cache = self.read_loss_cache(epoch, stage)
        return_values = []
        for key in keys:
            values = loss_cache[key].values
            if high is not None:
                assert num_values is not None, "num_values must be specified if high is specified"
                if high:
                    return_values.append(values[:num_values])
                elif not high:
                    return_values.append(values[-num_values:])
            else:
                return_values.append(values)
        return return_values

    def select_slides_for_epoch(
        self, epoch: int, stage: ModelKey, high: Optional[bool] = None, num_slides: Optional[int] = None
    ) -> np.ndarray:
        """Selects slides in ascending/descending order of loss values at a given epoch

        :param epoch: The epoch to select the slides from.
        :param stage: The model's stage (train/val).
        :param high: If True, selects the slides with the highest loss values, else selects the slides with the lowest
            loss values. If None, selects all slides.
        :param num_slides: The number of slides to select. If None, selects all slides.
        """
        return self._select_values_for_epoch([ResultsKey.SLIDE_ID], epoch, stage, high, num_slides)[0]

    def select_slides_entropy_for_epoch(
        self, epoch: int, stage: ModelKey, high: Optional[bool] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Selects slides and loss values of slides in ascending/descending order of loss at a given epoch

        :param epoch: The epoch to select the slides from.
        :param stage: The model's stage (train/val).
        :param high: If True, selects the slides with the highest loss values, else selects the slides with the lowest
            loss values. If None, selects all slides.
        """
        keys = [ResultsKey.SLIDE_ID, ResultsKey.ENTROPY]
        return_values = self._select_values_for_epoch(keys, epoch, stage, high, self.num_slides_scatter)
        return return_values[0], return_values[1]

    def select_all_losses_for_selected_slides(self, slides: np.ndarray, stage: ModelKey) -> LossDictType:
        """Selects the loss values for a given set of slides

        :param slides: The slides to select the loss values for.
        :param stage: The model's stage (train/val).
        """

        slides_loss_values: LossDictType = {slide_id: [] for slide_id in slides}
        for epoch in self.epochs_range:
            loss_cache = self.read_loss_cache(epoch, stage, idx_col=ResultsKey.SLIDE_ID)
            for slide_id in slides:
                slides_loss_values[slide_id].append(loss_cache.loc[slide_id, ResultsKey.LOSS])
        return slides_loss_values

    def select_slides_and_entropy_across_epochs(
        self, stage: ModelKey, high: bool = True
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Selects the slides with the highest/lowest loss values across epochs and their entropy values

        :param stage: The model's stage (train/val).
        :param high: If True, selects the slides with the highest loss values, else selects the slides with the lowest
        loss values.
        """
        slides = []
        slides_entropy = []
        for epoch in self.epochs_range:
            epoch_slides, epoch_slides_entropy = self.select_slides_entropy_for_epoch(epoch, stage, high)
            slides.append(epoch_slides)
            slides_entropy.append(epoch_slides_entropy)

        return np.array(slides).T, np.array(slides_entropy).T

    def save_slide_ids(self, slide_ids: List[str], path: Path) -> None:
        """Dumps the slides ids in a txt file.

        :param slide_ids: The slides ids to save.
        :param path: The path to save the slides ids to.
        """
        if slide_ids:
            with open(path, "w") as f:
                for slide_id in slide_ids:
                    f.write(f"{slide_id}\n")

    def sanity_check_loss_values(self, loss_values: LossDictType, stage: ModelKey) -> None:
        """Checks if there are any NaNs or any other potential annomalies in the loss values.

        :param loss_values: The loss values for all slides.
        :param stage: The model's stage (train/val).
        """
        # We don't want any of these exceptions to interrupt validation. So we catch them and log them.
        loss_values_copy = loss_values.copy()
        for slide_id, loss in loss_values_copy.items():
            try:
                if np.isnan(loss).any():
                    logging.warning(f"NaNs found in loss values for slide {slide_id}.")
                    self.nan_slides[stage].append(slide_id)
                    loss_values.pop(slide_id)
            except Exception as e:
                logging.warning(f"Error while checking for NaNs in loss values for slide {slide_id} with error {e}.")
                logging.warning(f"Loss values that caused the issue: {loss}")
                self.anomaly_slides[stage].append(slide_id)
                loss_values.pop(slide_id)
        self.save_slide_ids(self.nan_slides[stage], self.get_nan_slides_file(stage))
        self.save_slide_ids(self.anomaly_slides[stage], self.get_anomaly_slides_file(stage))

    def save_loss_ranks(self, slides_loss_values: LossDictType, stage: ModelKey) -> None:
        """Saves the loss ranks for each slide across all epochs and their respective statistics in csv files.

        :param slides_loss_values: The loss values for all slides.
        :param stage: The model's stage (train/val).
        """

        loss_df = pd.DataFrame(slides_loss_values).T
        loss_df.index.names = [ResultsKey.SLIDE_ID.value]
        loss_df.to_csv(self.get_all_epochs_loss_cache_file(stage))

        loss_stats = loss_df.T.describe().T.sort_values(by="mean", ascending=False)
        loss_stats.to_csv(self.get_loss_stats_file(stage))

        loss_ranks = loss_df.rank(ascending=False)
        loss_ranks.to_csv(self.get_loss_ranks_file(stage))

        loss_ranks_stats = loss_ranks.T.describe().T.sort_values("mean", ascending=True)
        loss_ranks_stats.to_csv(self.get_loss_ranks_stats_file(stage))

    def plot_slides_loss_scatter(
        self,
        slides: np.ndarray,
        slides_entropy: np.ndarray,
        stage: ModelKey,
        high: bool = True,
        figsize: Tuple[float, float] = (30, 30),
    ) -> None:
        """Plots the slides with the highest/lowest loss values across epochs in a scatter plot

        :param slides: The slides ids.
        :param slides_entropy: The entropy values for each slide.
        :param stage: The model's stage (train/val).
        :param figsize: The figure size, defaults to (30, 30)
        :param high: If True, plots the slides with the highest loss values, else plots the slides with the lowest loss.
        """
        label = self.TOP if high else self.BOTTOM
        plt.figure(figsize=figsize)
        markers_size = [15 * i for i in range(1, self.num_slides_scatter + 1)]
        markers_size = markers_size[::-1] if high else markers_size
        colors = cm.rainbow(np.linspace(0, 1, self.num_slides_scatter))
        for i in range(self.num_slides_scatter - 1, -1, -1):
            plt.scatter(self.epochs_range, slides[i], label=f"{label}_{i+1}", s=markers_size[i], color=colors[i])
            for entropy, epoch, slide in zip(slides_entropy[i], self.epochs_range, slides[i]):
                plt.annotate(f"{entropy:.3f}", (epoch, slide))
        plt.xlabel(self.X_LABEL)
        plt.ylabel(self.Y_LABEL)
        order = self.HIGHEST if high else self.LOWEST
        plt.title(f"Slides with {order} loss values per epoch and their entropy values")
        plt.xticks(self.epochs_range)
        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        plt.grid(True, linestyle="--")
        plt.savefig(self.get_scatter_plot_file(order, stage), bbox_inches="tight")

    def plot_loss_heatmap_for_slides_of_epoch(
        self,
        slides_loss_values: LossDictType,
        epoch: int,
        stage: ModelKey,
        high: bool,
        figsize: Tuple[float, float] = (15, 15)
    ) -> None:
        """Plots the loss values for each slide across all epochs in a heatmap.

        :param slides_loss_values: The loss values for each slide across all epochs.
        :param epoch: The epoch used to select the slides.
        :param stage: The model's stage (train/val).
        :param high: If True, plots the slides with the highest loss values, else plots the slides with the lowest loss.
        :param figsize: The figure size, defaults to (15, 15)
        """
        order = self.HIGHEST if high else self.LOWEST
        loss_values = np.array(list(slides_loss_values.values()))
        slides = list(slides_loss_values.keys())
        plt.figure(figsize=figsize)
        _ = sns.heatmap(loss_values, linewidth=0.5, annot=True, yticklabels=slides)
        plt.xlabel(self.X_LABEL)
        plt.ylabel(self.Y_LABEL)
        plt.title(f"Loss values evolution for {order} slides of epoch {epoch}")
        plt.savefig(self.get_heatmap_plot_file(epoch, order, stage), bbox_inches="tight")

    @staticmethod
    def compute_entropy(class_probs: torch.Tensor) -> List[float]:
        """Computes the entropy of the class probabilities.

        :param class_probs: The class probabilities.
        """
        return (-torch.sum(class_probs * torch.log(class_probs), dim=1)).tolist()

    def update_loss_cache(self, trainer: Trainer, outputs: BatchResultsType, batch: Dict, stage: ModelKey) -> None:
        """Updates the loss cache with the loss values for the current batch."""
        if self.should_cache_loss_values(trainer.current_epoch):
            self.loss_cache[stage][ResultsKey.LOSS].extend(outputs[ResultsKey.LOSS_PER_SAMPLE])
            self.loss_cache[stage][ResultsKey.SLIDE_ID].extend([slides[0] for slides in batch[ResultsKey.SLIDE_ID]])
            self.loss_cache[stage][ResultsKey.ENTROPY].extend(self.compute_entropy(outputs[ResultsKey.CLASS_PROBS]))
            if self.save_tile_ids:
                self.loss_cache[stage][ResultsKey.TILE_ID].extend(
                    [self.TILES_JOIN_TOKEN.join(tiles) for tiles in outputs[ResultsKey.TILE_ID]]
                )

    def synchronise_processes_and_reset(self, trainer: Trainer, pl_module: DeepMILModule, stage: ModelKey) -> None:
        """Synchronises the processes in DDP mode and resets the loss cache for the next epoch."""
        if self.should_cache_loss_values(trainer.current_epoch):
            self.gather_loss_cache(rank=pl_module.global_rank, stage=stage)
            if pl_module.global_rank == 0:
                self.save_loss_cache(trainer.current_epoch, stage)
            self.loss_cache[stage] = self.get_empty_loss_cache()  # reset loss cache for all processes

    def save_loss_outliers_analaysis_results(self, stage: ModelKey) -> None:
        """Saves the loss outliers analysis results."""
        all_slides = self.select_slides_for_epoch(epoch=0, stage=stage)
        all_loss_values_per_slides = self.select_all_losses_for_selected_slides(all_slides, stage=stage)

        self.sanity_check_loss_values(all_loss_values_per_slides, stage=stage)
        self.save_loss_ranks(all_loss_values_per_slides, stage=stage)

        top_slides, top_slides_entropy = self.select_slides_and_entropy_across_epochs(stage, high=True)
        self.plot_slides_loss_scatter(top_slides, top_slides_entropy, stage, high=True)

        bottom_slides, bottom_slides_entropy = self.select_slides_and_entropy_across_epochs(stage, high=False)
        self.plot_slides_loss_scatter(bottom_slides, bottom_slides_entropy, stage, high=False)

        for epoch in self.epochs_range:
            epoch_slides = self.select_slides_for_epoch(epoch, stage=stage)

            top_slides = epoch_slides[:self.num_slides_heatmap]
            top_slides_loss_values = self.select_all_losses_for_selected_slides(top_slides, stage=stage)
            self.plot_loss_heatmap_for_slides_of_epoch(top_slides_loss_values, epoch, stage, high=True)

            bottom_slides = epoch_slides[-self.num_slides_heatmap:]
            bottom_slides_loss_values = self.select_all_losses_for_selected_slides(bottom_slides, stage=stage)
            self.plot_loss_heatmap_for_slides_of_epoch(bottom_slides_loss_values, epoch, stage, high=False)
        self.loss_cache[stage] = self.get_empty_loss_cache()  # reset loss cache

    def handle_loss_exceptions(self, stage: ModelKey, exception: Exception) -> None:
        """Handles the loss exceptions. If log_exceptions is True, logs the exception as warnings, else raises it."""
        if self.log_exceptions:
            # If something goes wrong, we don't want to crash the training. We just log the error and carry on
            # validation.
            logging.warning(f"Error while detecting {stage} loss values outliers: {exception}")
        else:
            # If we want to debug the error, we raise it. This will crash the training. This is useful when
            # running smoke tests.
            raise Exception(f"Error while detecting {stage} loss values outliers: {exception}")

    def on_train_batch_end(  # type: ignore
        self,
        trainer: Trainer,
        pl_module: DeepMILModule,
        outputs: BatchResultsType,
        batch: Dict,
        batch_idx: int,
        unused: int = 0
    ) -> None:
        """Caches train loss values per slide at each training step in a local variable self.loss_cache."""
        self.update_loss_cache(trainer, outputs, batch, stage=ModelKey.TRAIN)

    def on_validation_batch_end(  # type: ignore
        self,
        trainer: Trainer,
        pl_module: DeepMILModule,
        outputs: BatchResultsType,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int
    ) -> None:
        """Caches validation loss values per slide at each training step in a local variable self.loss_cache."""
        self.update_loss_cache(trainer, outputs, batch, stage=ModelKey.VAL)

    def on_train_epoch_end(self, trainer: Trainer, pl_module: DeepMILModule) -> None:  # type: ignore
        """Gathers loss values per slide from all processes at the end of each epoch and saves them to a csv file."""
        self.synchronise_processes_and_reset(trainer, pl_module, ModelKey.TRAIN)

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: DeepMILModule) -> None:  # type: ignore
        """Gathers loss values per slide from all processes at the end of each epoch and saves them to a csv file."""
        self.synchronise_processes_and_reset(trainer, pl_module, ModelKey.VAL)

    def on_train_end(self, trainer: Trainer, pl_module: DeepMILModule) -> None:  # type: ignore
        """Hook called at the end of training. Plot the loss heatmap and scratter plots after ranking the slides by loss
        values."""
        if pl_module.global_rank == 0:
            try:
                self.save_loss_outliers_analaysis_results(stage=ModelKey.TRAIN)
            except Exception as e:
                self.handle_loss_exceptions(stage=ModelKey.TRAIN, exception=e)

    def on_validation_end(self, trainer: Trainer, pl_module: DeepMILModule) -> None:  # type: ignore
        """Hook called at the end of validation. Plot the loss heatmap and scratter plots after ranking the slides by
        loss values."""
        epoch = trainer.current_epoch
        if pl_module.global_rank == 0 and not pl_module._on_extra_val_epoch and epoch == (self.max_epochs - 1):
            try:
                self.save_loss_outliers_analaysis_results(stage=ModelKey.VAL)
            except Exception as e:
                self.handle_loss_exceptions(stage=ModelKey.VAL, exception=e)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import logging
from pathlib import Path
from typing import Any, Collection, List, Optional, Sequence, Tuple, Dict

from sklearn.metrics import confusion_matrix
from torch import Tensor
import matplotlib.pyplot as plt

from health_cpath.datasets.base_dataset import SlidesDataset
from health_cpath.preprocessing.loading import LoadingParams
from health_cpath.utils.viz_utils import (
    plot_attention_tiles,
    plot_heatmap_overlay,
    plot_attention_histogram,
    plot_normalized_confusion_matrix,
    plot_scores_hist,
    plot_slide,
)
from health_cpath.utils.analysis_plot_utils import plot_pr_curve, format_pr_or_roc_axes
from health_cpath.utils.naming import PlotOption, ResultsKey, SlideKey
from health_cpath.utils.tiles_selection_utils import SlideNode, TilesSelector
from health_cpath.utils.viz_utils import save_figure


ResultsType = Dict[ResultsKey, List[Any]]
SlideDictType = Dict[SlideKey, Any]


def validate_class_names_for_plot_options(
    class_names: Optional[Sequence[str]], plot_options: Collection[PlotOption]
) -> Optional[Sequence[str]]:
    """Make sure that class names are provided if `PlotOption.CONFUSION_MATRIX` is among the chosen plot_options."""
    if PlotOption.CONFUSION_MATRIX in plot_options and not class_names:
        raise ValueError(
            "No class_names were provided while activating confusion matrix plotting. You need to specify the class "
            "names to use the `PlotOption.CONFUSION_MATRIX`"
        )
    return class_names


def save_scores_histogram(results: ResultsType, figures_dir: Path, stage: str = '') -> None:
    """Plots and saves histogram scores figure in its dedicated directory.

    :param results: Dict of lists that contains slide_level results
    :param figures_dir: The path to the directory where to save the histogram scores.
    :param stage: Test or validation, used to name the figure. Empty string by default.
    """
    fig = plot_scores_hist(results)
    save_figure(fig=fig, figpath=figures_dir / f"hist_scores_{stage}.png")


def save_pr_curve(results: ResultsType, figures_dir: Path, stage: str = '') -> None:
    """Plots and saves PR curve figure in its dedicated directory. This implementation
    only works for binary classification.
''
    :param results: Dict of lists that contains slide_level results
    :param figures_dir: The path to the directory where to save the histogram scores
    :param stage: Test or validation, used to name the figure. Empty string by default.
    """
    true_labels = [i.item() if isinstance(i, Tensor) else i for i in results[ResultsKey.TRUE_LABEL]]
    if len(set(true_labels)) == 2:
        scores = [i.item() if isinstance(i, Tensor) else i for i in results[ResultsKey.PROB]]
        fig, ax = plt.subplots()
        plot_pr_curve(true_labels, scores, legend_label=stage, ax=ax)
        ax.legend()
        format_pr_or_roc_axes(plot_type='pr', ax=ax)
        save_figure(fig=fig, figpath=figures_dir / f"pr_curve_{stage}.png")
    else:
        logging.warning("The PR curve plot implementation works only for binary cases, this plot will be skipped.")


def save_confusion_matrix(results: ResultsType, class_names: Sequence[str], figures_dir: Path, stage: str = '') -> None:
    """Plots and saves confusion matrix figure in its dedicated directory.

    :param results: Dict of lists that contains slide_level results
    :param class_names: List of class names.
    :param figures_dir: The path to the directory where to save the confusion matrix.
    :param stage: Test or validation, used to name the figure. Empty string by default.
    """
    true_labels = [i.item() if isinstance(i, Tensor) else i for i in results[ResultsKey.TRUE_LABEL]]
    pred_labels = [i.item() if isinstance(i, Tensor) else i for i in results[ResultsKey.PRED_LABEL]]
    all_potential_labels = list(range(len(class_names)))
    true_labels_diff_expected = set(true_labels).difference(set(all_potential_labels))
    pred_labels_diff_expected = set(pred_labels).difference(set(all_potential_labels))

    if true_labels_diff_expected != set():
        raise ValueError("More entries were found in true labels than are available in class names")
    if pred_labels_diff_expected != set():
        raise ValueError("More entries were found in predicted labels than are available in class names")

    cf_matrix_n = confusion_matrix(
        true_labels,
        pred_labels,
        labels=all_potential_labels,
        normalize="true"
    )

    fig = plot_normalized_confusion_matrix(cm=cf_matrix_n, class_names=(class_names))
    save_figure(fig=fig, figpath=figures_dir / f"normalized_confusion_matrix_{stage}.png")


def save_top_and_bottom_tiles(
    case: str, slide_node: SlideNode, figures_dir: Path, num_columns: int = 4, figsize: Tuple[int, int] = (10, 10)
) -> None:
    """Plots and saves the top and bottom attention tiles of a given slide_node

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: the slide_node for which we plot top and bottom tiles.
    :param figures_dir: The path to the directory where to save the attention tiles figure.
    """
    top_tiles_fig = plot_attention_tiles(
        slide_node=slide_node, case=case, top=True, num_columns=num_columns, figsize=figsize
    )
    save_figure(fig=top_tiles_fig, figpath=figures_dir / f"{slide_node.slide_id}_top.png")

    bottom_tiles_fig = plot_attention_tiles(
        slide_node=slide_node, case=case, top=False, num_columns=num_columns, figsize=figsize
    )
    save_figure(fig=bottom_tiles_fig, figpath=figures_dir / f"{slide_node.slide_id}_bottom.png")


def save_slide_thumbnail(case: str, slide_node: SlideNode, slide_dict: SlideDictType, figures_dir: Path) -> None:
    """Plots and saves a slide thumbnail

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: The slide node that encapsulates the slide metadata.
    :param slide_dict: The slide dictionary that contains the slide image and other metadata.
    :param figures_dir: The path to the directory where to save the plots.
    """
    fig = plot_slide(case=case, slide_node=slide_node, slide_image=slide_dict[SlideKey.IMAGE], scale=1.0)
    save_figure(fig=fig, figpath=figures_dir / f"{slide_node.slide_id}_thumbnail.png")


def save_attention_heatmap(
    case: str,
    slide_node: SlideNode,
    slide_dict: SlideDictType,
    figures_dir: Path,
    results: ResultsType,
    tile_size: int = 224,
    should_upscale_coords: bool = False,
) -> None:
    """Plots and saves a slide thumbnail and attention heatmap

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: The slide node that encapsulates the slide metadata.
    :param slide_dict: The slide dictionary that contains the slide image and other metadata.
    :param figures_dir: The path to the directory where to save the plots.
    :param results: Dict containing ResultsKey keys (e.g. slide id) and values as lists of output slides.
    :param tile_size: Size of each tile. Default 224.
    :param should_upscale_coords: Whether to upscale the coordinates of the attention heatmap. Default False.
    """
    fig = plot_heatmap_overlay(
        case=case,
        slide_node=slide_node,
        slide_dict=slide_dict,
        results=results,
        tile_size=tile_size,
        should_upscale_coords=should_upscale_coords,
    )
    save_figure(fig=fig, figpath=figures_dir / f"{slide_node.slide_id}_heatmap.png")


def save_attention_histogram(case: str, slide_node: SlideNode, results: ResultsType, figures_dir: Path) -> None:
    """Plots a histogram of the attention values of the tiles in a bag.

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: The slide node that encapsulates the slide metadata.
    :param results: Dict containing ResultsKey keys (e.g. slide id) and values as lists of output slides.
    :param figures_dir: The path to the directory where to save the plot.
    """
    fig = plot_attention_histogram(case=case, slide_node=slide_node, results=results)
    save_figure(fig=fig, figpath=figures_dir / f"{slide_node.slide_id}_histogram.png")


def make_figure_dirs(subfolder: str, parent_dir: Path) -> Path:
    """Create the figure directory"""
    figures_dir = parent_dir / subfolder
    figures_dir.mkdir(parents=True, exist_ok=True)
    return figures_dir


class DeepMILPlotsHandler:
    def __init__(
        self,
        plot_options: Collection[PlotOption],
        loading_params: LoadingParams,
        tile_size: int = 224,
        num_columns: int = 4,
        figsize: Tuple[int, int] = (10, 10),
        stage: str = '',
        class_names: Optional[Sequence[str]] = None,
    ) -> None:
        """Class that handles the plotting of DeepMIL results.

        :param plot_options: A set of plot options to produce the desired plot outputs.
        :param loading_params: The loading parameters to use when loading the whole slide images.
        :param tile_size: The size of the tiles to use when plotting the attention tiles, defaults to 224
        :param num_columns: Number of columns to create the subfigures grid, defaults to 4
        :param figsize: The figure size of tiles attention plots, defaults to (10, 10)
        :param stage: Test or Validation, used to name the plots
        :param class_names: List of class names, defaults to None
        """

        self.plot_options = plot_options
        self.class_names = validate_class_names_for_plot_options(class_names, plot_options)
        self.tile_size = tile_size
        self.num_columns = num_columns
        self.figsize = figsize
        self.stage = stage
        self.loading_params = loading_params
        self.should_upscale_coords = loading_params.should_upscale_coordinates()
        self.loading_params.set_roi_type_to_foreground()
        self.slides_dataset: Optional[SlidesDataset] = None

    def get_slide_dict(self, slide_node: SlideNode) -> SlideDictType:
        """Returns the slide dictionary for a given slide node"""
        assert self.slides_dataset is not None, "Cannot plot attention heatmap or wsi without slides dataset"
        slide_index = self.slides_dataset.dataset_df.index.get_loc(slide_node.slide_id)
        assert isinstance(slide_index, int), f"Got non-unique slide ID: {slide_node.slide_id}"
        slide_dict = self.slides_dataset[slide_index]
        loader = self.loading_params.get_load_roid_transform()
        slide_dict = loader(slide_dict)
        return slide_dict

    def save_slide_node_figures(
        self, case: str, slide_node: SlideNode, outputs_dir: Path, results: ResultsType
    ) -> None:
        """Plots and saves all slide related figures: `TOP_BOTTOM_TILES`, `SLIDE_THUMBNAIL` and `ATTENTION_HEATMAP`."""
        case_dir = make_figure_dirs(subfolder=case, parent_dir=outputs_dir)

        if PlotOption.TOP_BOTTOM_TILES in self.plot_options:
            save_top_and_bottom_tiles(case, slide_node, case_dir, self.num_columns, self.figsize)

        if PlotOption.ATTENTION_HISTOGRAM in self.plot_options:
            save_attention_histogram(case, slide_node, results, case_dir)

        if PlotOption.ATTENTION_HEATMAP in self.plot_options or PlotOption.SLIDE_THUMBNAIL in self.plot_options:
            slide_dict = self.get_slide_dict(slide_node=slide_node)

            if PlotOption.SLIDE_THUMBNAIL in self.plot_options:
                save_slide_thumbnail(case=case, slide_node=slide_node, slide_dict=slide_dict, figures_dir=case_dir)

            if PlotOption.ATTENTION_HEATMAP in self.plot_options:
                save_attention_heatmap(
                    case, slide_node, slide_dict, case_dir, results, self.tile_size, self.should_upscale_coords
                )

    def save_plots(self, outputs_dir: Path, tiles_selector: Optional[TilesSelector], results: ResultsType) -> None:
        """Plots and saves all selected plot options during inference (validation or test) time.

        :param outputs_dir: The root output directory where to save plots figures.
        :param tiles_selector: The tiles selector used to select top and bottom tiles from top and bottom slides.
        :param results: A dictionary of the validation or tests results.
        :param stage: The model stage: validation or test.
        """
        if self.plot_options:
            logging.info(
                f"Plotting {[opt.value for opt in self.plot_options]}..."
            )
            figures_dir = make_figure_dirs(subfolder="fig", parent_dir=outputs_dir)

            if PlotOption.PR_CURVE in self.plot_options:
                save_pr_curve(results=results, figures_dir=figures_dir, stage=self.stage)

            if PlotOption.HISTOGRAM in self.plot_options:
                save_scores_histogram(results=results, figures_dir=figures_dir, stage=self.stage,)

            if PlotOption.CONFUSION_MATRIX in self.plot_options:
                assert self.class_names
                save_confusion_matrix(results, class_names=self.class_names, figures_dir=figures_dir, stage=self.stage)

            if tiles_selector:
                for class_id in range(tiles_selector.n_classes):

                    for slide_node in tiles_selector.top_slides_heaps[class_id]:
                        case = "TN" if class_id == 0 else f"TP_{class_id}"
                        self.save_slide_node_figures(case, slide_node, outputs_dir, results)

                    for slide_node in tiles_selector.bottom_slides_heaps[class_id]:
                        case = "FP" if class_id == 0 else f"FN_{class_id}"
                        self.save_slide_node_figures(case, slide_node, outputs_dir, results)



# TODO: remove innereye references
innereye_ssl_checkpoint = "hsharma_panda_explore_1638437076_357167ae"
innereye_ssl_checkpoint_binary = "PANDA_SimCLR_1654023538_7a0f1d00"
innereye_ssl_checkpoint_crck_4ws = "CRCK_SimCLR_1653673515_42d53d78"
innereye_ssl_checkpoint_crck_radiomics = "704b1af8-7c75-46ed-8460-d80a0e603194"

# outdated checkpoints
# innereye_ssl_checkpoint_crck_radiomics = updated_transforms:updated_transforms_1636471522_5473e3ff



from azure.storage.blob import generate_blob_sas, BlobSasPermissions
from azureml.core import Workspace
from datetime import datetime, timedelta
from health_azure import get_workspace
from health_ml.utils.common_utils import DEFAULT_AML_CHECKPOINT_DIR
from typing import Optional


def get_checkpoint_url_from_aml_run(
    run_id: str,
    checkpoint_filename: str,
    expiry_days: int = 1,
    aml_workspace: Optional[Workspace] = None,
    sas_token: Optional[str] = None,
) -> str:
    """Generate a SAS URL for the checkpoint file in the given run.

    :param run_id: The run ID of the checkpoint.
    :param checkpoint_filename: The filename of the checkpoint.
    :param expiry_days: The number of days the SAS URL is valid for, defaults to 30.
    :param aml_workspace: The Azure ML workspace to use, defaults to the default workspace.
    :param sas_token: The SAS token to use, defaults to None.
    :return: The SAS URL for the checkpoint.
    """
    datastore = get_workspace(aml_workspace=aml_workspace).get_default_datastore()
    account_name = datastore.account_name
    container_name = 'azureml'
    blob_name = f'ExperimentRun/dcid.{run_id}/{DEFAULT_AML_CHECKPOINT_DIR}/{checkpoint_filename}'

    if not sas_token:
        sas_token = generate_blob_sas(account_name=datastore.account_name,
                                      container_name=container_name,
                                      blob_name=blob_name,
                                      account_key=datastore.account_key,
                                      permission=BlobSasPermissions(read=True),
                                      expiry=datetime.utcnow() + timedelta(days=expiry_days))

    return f'https://{account_name}.blob.core.windows.net/{container_name}/{blob_name}?{sas_token}'


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--run_id', type=str, help='The run id of the model checkpoint')
    parser.add_argument('--checkpoint_filename', type=str, default='last.ckpt',
                        help='The filename of the model checkpoint. Default: last.ckpt')
    parser.add_argument('--expiry_days', type=int, default=30,
                        help='The number of hours for which the SAS token is valid. Default: 30 for 1 month')
    args = parser.parse_args()
    url = get_checkpoint_url_from_aml_run(args.run_id, args.checkpoint_filename, args.expiry_days)
    print(f'Checkpoint URL: {url}')



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

PANDA_DATASET_ID = "panda"
PANDA_5X_TILES_DATASET_ID = "PANDA_5X_tiles_20220714_134233_level1_224"
PANDA_20X_TILES_DATASET_ID = "PANDA_20X_tiles_20220714_142937_level0_224"

TCGA_CRCK_DATASET_ID = "TCGA-CRCk"

TCGA_PRAD_DATASET_ID = "TCGA-PRAD_20220712"
TCGA_PRAD_10X_TILES_DATASET_ID = "TCGA-PRAD_10X_tiles_20220728_123811_level1_224"

DEFAULT_DATASET_LOCATION = "/tmp/datasets/"
PANDA_5X_TILES_DATASET_DIR = DEFAULT_DATASET_LOCATION + PANDA_5X_TILES_DATASET_ID
PANDA_20X_TILES_DATASET_DIR = DEFAULT_DATASET_LOCATION + PANDA_20X_TILES_DATASET_ID
TCGA_CRCK_DATASET_DIR = DEFAULT_DATASET_LOCATION + TCGA_CRCK_DATASET_ID
TCGA_PRAD_DATASET_DIR = DEFAULT_DATASET_LOCATION + TCGA_PRAD_DATASET_ID
TCGA_PRAD_10X_TILES_DATASET_DIR = DEFAULT_DATASET_LOCATION + TCGA_PRAD_DATASET_ID
